# -*- test-case-name: xmantissa.test.test_sharing -*-

"""
This module provides various abstractions for sharing public data in Axiom.
"""

import os
import warnings

from zope.interface import implementedBy, directlyProvides, Interface

from twisted.python.reflect import qual, namedAny
from twisted.protocols.amp import Argument, Box, parseString

from epsilon.structlike import record

from axiom import userbase
from axiom.item import Item
from axiom.attributes import reference, text, AND
from axiom.upgrade import registerUpgrader


ALL_IMPLEMENTED_DB = u'*'
ALL_IMPLEMENTED = object()


class NoSuchShare(Exception):
    """
    User requested an object that doesn't exist, was not allowed.
    """



class ConflictingNames(Exception):
    """
    The same name was defined in two separate interfaces.
    """



class RoleRelationship(Item):
    """
    RoleRelationship is a bridge record linking member roles with group roles
    that they are members of.
    """
    schemaVersion = 1
    typeName = 'sharing_relationship'

    member = reference(
        doc="""
        This is a reference to a L{Role} which is a member of my 'group' attribute.
        """)

    group = reference(
        doc="""
        This is a reference to a L{Role} which represents a group that my 'member'
        attribute is a member of.
        """)


def _entuple(r):
    """
    Convert a L{record} to a tuple.
    """
    return tuple(getattr(r, n) for n in r.__names__)



class Identifier(record('shareID localpart domain')):
    """
    A fully-qualified identifier for an entity that can participate in a
    message either as a sender or a receiver.
    """

    @classmethod
    def fromSharedItem(cls, sharedItem):
        """
        Return an instance of C{cls} derived from the given L{Item} that has
        been shared.

        Note that this API does not provide any guarantees of which result it
        will choose.  If there are are multiple possible return values, it will
        select and return only one.  Items may be shared under multiple
        L{shareID}s.  A user may have multiple valid account names.  It is
        sometimes impossible to tell from context which one is appropriate, so
        if your application has another way to select a specific shareID you
        should use that instead.

        @param sharedItem: an L{Item} that should be shared.

        @return: an L{Identifier} describing the C{sharedItem} parameter.

        @raise L{NoSuchShare}: if the given item is not shared or its store
        does not contain any L{LoginMethod} items which would identify a user.
        """
        localpart = None
        for (localpart, domain) in userbase.getAccountNames(sharedItem.store):
            break
        if localpart is None:
            raise NoSuchShare()
        for share in sharedItem.store.query(Share,
                                            Share.sharedItem == sharedItem):
            break
        else:
            raise NoSuchShare()
        return cls(
            shareID=share.shareID,
            localpart=localpart, domain=domain)


    def __cmp__(self, other):
        """
        Compare this L{Identifier} to another object.
        """
        # Note - might be useful to have this usable by arbitrary L{record}
        # objects.  It can't be the default, but perhaps a mixin?
        if not isinstance(other, Identifier):
            return NotImplemented
        return cmp(_entuple(self), _entuple(other))



class IdentifierArgument(Argument):
    """
    An AMP argument which can serialize and deserialize an L{Identifier}.
    """

    def toString(self, obj):
        """
        Convert the given L{Identifier} to a string.
        """
        return Box(shareID=obj.shareID.encode('utf-8'),
                   localpart=obj.localpart.encode('utf-8'),
                   domain=obj.domain.encode('utf-8')).serialize()


    def fromString(self, inString):
        """
        Convert the given string to an L{Identifier}.
        """
        box = parseString(inString)[0]
        return Identifier(shareID=box['shareID'].decode('utf-8'),
                          localpart=box['localpart'].decode('utf-8'),
                          domain=box['domain'].decode('utf-8'))



class Role(Item):
    """
    A Role is an identifier for a group or individual which has certain
    permissions.

    Items shared within the sharing system are always shared with a particular
    role.
    """

    schemaVersion = 1
    typeName = 'sharing_role'

    externalID = text(
        doc="""
        This is the external identifier which the role is known by.  This field is
        used to associate users with their primary role.  If a user logs in as
        bob@divmod.com, the sharing system will associate his primary role with
        the pre-existing role with the externalID of 'bob@divmod.com', or
        'Everybody' if no such role exists.

        For group roles, the externalID is not currently used except as a
        display identifier.  Group roles should never have an '@' character in
        them, however, to avoid confusion with user roles.
        """, allowNone=False)

    # XXX TODO: In addition to the externalID, we really need to have something
    # that identifies what protocol the user for the role is expected to log in
    # as, and a way to identify the way that their role was associated with
    # their login.  For example, it might be acceptable for some security
    # applications (e.g. spam prevention) to simply use an HTTP cookie.  For
    # others (accounting database manipulation) it should be possible to
    # require more secure methods of authentication, like a signed client
    # certificate.

    description = text(
        doc="""
        This is a free-form descriptive string for use by users to explain the
        purpose of the role.  Since the externalID is used by security logic
        and must correspond to a login identifier, this can be used to hold a
        user's real name.
        """)


    def becomeMemberOf(self, groupRole):
        """
        Instruct this (user or group) Role to become a member of a group role.

        @param groupRole: The role that this group should become a member of.
        """
        self.store.findOrCreate(RoleRelationship,
                                group=groupRole,
                                member=self)


    def allRoles(self, memo=None):
        """
        Identify all the roles that this role is authorized to act as.

        @param memo: used only for recursion.  Do not pass this.

        @return: an iterator of all roles that this role is a member of,
        including itself.
        """
        if memo is None:
            memo = set()
        elif self in memo:
            # this is bad, but we have successfully detected and prevented the
            # only really bad symptom, an infinite loop.
            return
        memo.add(self)
        yield self
        for groupRole in self.store.query(Role,
                                          AND(RoleRelationship.member == self,
                                              RoleRelationship.group == Role.storeID)):
            for roleRole in groupRole.allRoles(memo):
                yield roleRole


    def shareItem(self, sharedItem, shareID=None, interfaces=ALL_IMPLEMENTED):
        """
        Share an item with this role.  This provides a way to expose items to
        users for later retrieval with L{Role.getShare}.

        @param sharedItem: an item to be shared.

        @param shareID: a unicode string.  If provided, specify the ID under which
        the shared item will be shared.

        @param interfaces: a list of Interface objects which specify the methods
        and attributes accessible to C{toRole} on C{sharedItem}.

        @return: a L{Share} which records the ability of the given role to
        access the given item.
        """
        if shareID is None:
            shareID = genShareID(sharedItem.store)
        return Share(store=self.store,
                     shareID=shareID,
                     sharedItem=sharedItem,
                     sharedTo=self,
                     sharedInterfaces=interfaces)


    def getShare(self, shareID):
        """
        Retrieve a proxy object for a given shareID, previously shared with
        this role or one of its group roles via L{Role.shareItem}.

        @return: a L{SharedProxy}.  This is a wrapper around the shared item
        which only exposes those interfaces explicitly allowed for the given
        role.

        @raise: L{NoSuchShare} if there is no item shared to the given role for
        the given shareID.
        """
        shares = list(
            self.store.query(Share,
                             AND(Share.shareID == shareID,
                                 Share.sharedTo.oneOf(self.allRoles()))))
        interfaces = []
        for share in shares:
            interfaces += share.sharedInterfaces
        if shares:
            return SharedProxy(shares[0].sharedItem,
                               interfaces,
                               shareID)
        raise NoSuchShare()


    def asAccessibleTo(self, query):
        """
        @param query: An Axiom query describing the Items to retrieve, which this
        role can access.
        @type query: an L{iaxiom.IQuery} provider.

        @return: an iterable which yields the shared proxies that are available
        to the given role, from the given query.
        """
        # XXX TODO #2371: this method really *should* be returning an L{IQuery}
        # provider as well, but that is kind of tricky to do.  Currently, doing
        # queries leaks authority, because the resulting objects have stores
        # and "real" items as part of their interface; having this be a "real"
        # query provider would obviate the need to escape the L{SharedProxy}
        # security constraints in order to do any querying.
        allRoles = list(self.allRoles())
        count = 0
        unlimited = query.cloneQuery(limit=None)
        for result in unlimited:
            allShares = list(query.store.query(
                    Share,
                    AND(Share.sharedItem == result,
                        Share.sharedTo.oneOf(allRoles))))
            interfaces = []
            for share in allShares:
                interfaces += share.sharedInterfaces
            if allShares:
                count += 1
                yield SharedProxy(result, interfaces, allShares[0].shareID)
                if count == query.limit:
                    return




class _really(object):
    """
    A dynamic proxy for dealing with 'private' attributes on L{SharedProxy},
    which overrides C{__getattribute__} itself.  This is pretty syntax to avoid
    ugly references to __dict__ and super and object.__getattribute__() in
    dynamic proxy implementations.
    """

    def __init__(self, orig):
        """
        Create a _really object with a dynamic proxy.

        @param orig: an object that overrides __getattribute__, probably
        L{SharedProxy}.
        """
        self.orig = orig


    def __setattr__(self, name, value):
        """
        Set an attribute on my original, unless my original has not yet been set,
        in which case set it on me.
        """
        try:
            orig = object.__getattribute__(self, 'orig')
        except AttributeError:
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(orig, name, value)


    def __getattribute__(self, name):
        """
        Get an attribute present on my original using L{object.__getattribute__},
        not the overridden version.
        """
        return object.__getattribute__(object.__getattribute__(self, 'orig'),
                                       name)



ALLOWED_ON_PROXY = ['__provides__', '__dict__']

class SharedProxy(object):
    """
    A shared proxy is a dynamic proxy which provides exposes methods and
    attributes declared by shared interfaces on a given Item.  These are
    returned from L{Role.getShare} and yielded from L{Role.asAccessibleTo}.

    Shared proxies are unlike regular items because they do not have 'storeID'
    or 'store' attributes (unless explicitly exposed).  They are designed
    to make security easy to implement: if you have a shared proxy, you can
    access any attribute or method on it without having to do explicit
    permission checks.

    If you *do* want to perform an explicit permission check, for example, to
    render some UI associated with a particular permission, it can be performed
    as a functionality check instead.  For example, C{if getattr(proxy,
    'feature', None) is None:} or, more formally,
    C{IFeature.providedBy(proxy)}.  If your interfaces are all declared and
    implemented properly everywhere, these checks will work both with shared
    proxies and with the original Items that they represent (but of course, the
    original Items will always provide all of their features).

    (Note that object.__getattribute__ still lets you reach inside any object,
    so don't imagine this makes you bulletproof -- you have to cooperate with
    it.)
    """

    def __init__(self, sharedItem, sharedInterfaces, shareID):
        """
        Create a shared proxy for a given item.

        @param sharedItem: The original item that was shared.

        @param sharedInterfaces: a list of interfaces which C{sharedItem}
        implements that this proxy should allow access to.

        @param shareID: the external identifier that the shared item was shared
        as.
        """
        rself = _really(self)
        rself._sharedItem = sharedItem
        rself._shareID = shareID
        rself._adapterCache = {}
        # Drop all duplicate shared interfaces.
        uniqueInterfaces = list(sharedInterfaces)
        # XXX there _MUST_ Be a better algorithm for this
        for left in sharedInterfaces:
            for right in sharedInterfaces:
                if left.extends(right) and right in uniqueInterfaces:
                    uniqueInterfaces.remove(right)
        for eachInterface in uniqueInterfaces:
            if not eachInterface.providedBy(sharedItem):
                impl = eachInterface(sharedItem, None)
                if impl is not None:
                    rself._adapterCache[eachInterface] = impl
        rself._sharedInterfaces = uniqueInterfaces
        # Make me look *exactly* like the item I am proxying for, at least for
        # the purposes of adaptation
        # directlyProvides(self, providedBy(sharedItem))
        directlyProvides(self, uniqueInterfaces)


    def __repr__(self):
        """
        Return a pretty string representation of this shared proxy.
        """
        rself = _really(self)
        return 'SharedProxy(%r, %r, %r)' % (
            rself._sharedItem,
            rself._sharedInterfaces,
            rself._shareID)


    def __getattribute__(self, name):
        """
        @return: attributes from my shared item, present in the shared interfaces
        list for this proxy.

        @param name: the name of the attribute to retrieve.

        @raise AttributeError: if the attribute was not found or access to it was
        denied.
        """
        if name in ALLOWED_ON_PROXY:
            return object.__getattribute__(self, name)
        rself = _really(self)
        if name == 'sharedInterfaces':
            return rself._sharedInterfaces
        elif name == 'shareID':
            return rself._shareID
        for iface in rself._sharedInterfaces:
            if name in iface:
                if iface in rself._adapterCache:
                    return getattr(rself._adapterCache[iface], name)
                return getattr(rself._sharedItem, name)
        raise AttributeError("%r has no attribute %r" % (self, name))


    def __setattr__(self, name, value):
        """
        Set an attribute on the shared item.  If the name of the attribute is in
        L{ALLOWED_ON_PROXY}, set it on this proxy instead.

        @param name: the name of the attribute to set

        @param value: the value of the attribute to set

        @return: None
        """
        if name in ALLOWED_ON_PROXY:
            self.__dict__[name] = value
        else:
            raise AttributeError("unsettable: "+repr(name))



def _interfacesToNames(interfaces):
    """
    Convert from a list of interfaces to a unicode string of names suitable for
    storage in the database.

    @param interfaces: an iterable of Interface objects.

    @return: a unicode string, a comma-separated list of names of interfaces.

    @raise ConflictingNames: if any of the names conflict: see
    L{_checkConflictingNames}.
    """
    if interfaces is ALL_IMPLEMENTED:
        names = ALL_IMPLEMENTED_DB
    else:
        _checkConflictingNames(interfaces)
        names = u','.join(map(qual, interfaces))
    return names



class Share(Item):
    """
    A Share is a declaration that users with a given role can access a given
    set of functionality, as described by an Interface object.

    They should be created with L{Role.shareItem} and retrieved with
    L{Role.asAccessibleTo} and L{Role.getShare}.
    """

    schemaVersion = 2
    typeName = 'sharing_share'

    shareID = text(
        doc="""
        The shareID is the externally-visible identifier for this share.  It is
        free-form text, which users may enter to access this share.

        Currently the only concrete use of this attribute is in HTTP[S] URLs, but
        in the future it will be used in menu entries.
        """,
        allowNone=False)

    sharedItem = reference(
        doc="""
        The sharedItem attribute is a reference to the item which is being
        provided.
        """,
        allowNone=False,
        whenDeleted=reference.CASCADE)

    sharedTo = reference(
        doc="""
        The sharedTo attribute is a reference to the Role which this item is shared
        with.
        """,
        allowNone=False)

    sharedInterfaceNames = text(
        doc="""
        This is an internal implementation detail of the sharedInterfaces
        attribute.
        """,
        allowNone=False)


    def __init__(self, **kw):
        """
        Create a share.

        Consider this interface private; use L{shareItem} instead.
        """
        # XXX TODO: All I really want to do here is to have enforcement of
        # allowNone happen at the _end_ of __init__; axiom should probably do
        # that by default, since there are several __init__s like this which
        # don't really do anything scattered throughout the codebase.
        kw['sharedInterfaceNames'] = _interfacesToNames(kw.pop('sharedInterfaces'))
        super(Share, self).__init__(**kw)


    def sharedInterfaces():
        """
        This attribute is the public interface for code which wishes to discover
        the list of interfaces allowed by this Share.  It is a list of
        Interface objects.
        """
        def get(self):
            if not self.sharedInterfaceNames:
                return ()
            if self.sharedInterfaceNames == ALL_IMPLEMENTED_DB:
                I = implementedBy(self.sharedItem.__class__)
                L = list(I)
                T = tuple(L)
                return T
            else:
                return tuple(map(namedAny, self.sharedInterfaceNames.split(u',')))
        def set(self, newValue):
            self.sharedAttributeNames = _interfacesToNames(newValue)
        return get, set

    sharedInterfaces = property(
        doc=sharedInterfaces.__doc__,
        *sharedInterfaces())



def upgradeShare1to2(oldShare):
    "Upgrader from Share version 1 to version 2."
    sharedInterfaces = []
    attrs = set(oldShare.sharedAttributeNames.split(u','))
    for iface in implementedBy(oldShare.sharedItem.__class__):
        if set(iface) == attrs or attrs == set('*'):
            sharedInterfaces.append(iface)

    newShare = oldShare.upgradeVersion('sharing_share', 1, 2,
                                       shareID=oldShare.shareID,
                                       sharedItem=oldShare.sharedItem,
                                       sharedTo=oldShare.sharedTo,
                                       sharedInterfaces=sharedInterfaces)
    return newShare


registerUpgrader(upgradeShare1to2, 'sharing_share', 1, 2)



def genShareID(store):
    """
    Generate a new, randomized share-ID for use as the default of shareItem, if
    none is specified.

    @return: a random share-ID.

    @rtype: unicode.
    """
    return unicode(os.urandom(16).encode('hex'), 'ascii')



def getEveryoneRole(store):
    """
    Get a base 'Everyone' role for this store, which is the role that every
    user, including the anonymous user, has.
    """
    return store.findOrCreate(Role, externalID=u'Everyone')



def getAuthenticatedRole(store):
    """
    Get the base 'Authenticated' role for this store, which is the role that is
    given to every user who is explicitly identified by a non-anonymous
    username.
    """
    def tx():
        def addToEveryone(newAuthenticatedRole):
            newAuthenticatedRole.becomeMemberOf(getEveryoneRole(store))
            return newAuthenticatedRole
        return store.findOrCreate(Role, addToEveryone, externalID=u'Authenticated')
    return store.transact(tx)



def getPrimaryRole(store, primaryRoleName, createIfNotFound=False):
    """
    Get Role object corresponding to an identifier name.  If the role name
    passed is the empty string, it is assumed that the user is not
    authenticated, and the 'Everybody' role is primary.  If the role name
    passed is non-empty, but has no corresponding role, the 'Authenticated'
    role - which is a member of 'Everybody' - is primary.  Finally, a specific
    role can be primary if one exists for the user's given credentials, that
    will automatically always be a member of 'Authenticated', and by extension,
    of 'Everybody'.

    @param primaryRoleName: a unicode string identifying the role to be
    retrieved.  This corresponds to L{Role}'s externalID attribute.

    @param createIfNotFound: a boolean.  If True, create a role for the given
    primary role name if no exact match is found.  The default, False, will
    instead retrieve the 'nearest match' role, which can be Authenticated or
    Everybody depending on whether the user is logged in or not.

    @return: a L{Role}.
    """
    if not primaryRoleName:
        return getEveryoneRole(store)
    ff = store.findUnique(Role, Role.externalID == primaryRoleName, default=None)
    if ff is not None:
        return ff
    authRole = getAuthenticatedRole(store)
    if createIfNotFound:
        role = Role(store=store,
                    externalID=primaryRoleName)
        role.becomeMemberOf(authRole)
        return role
    return authRole



def getSelfRole(store):
    """
    Retrieve the Role which corresponds to the user to whom the given store
    belongs.
    """
    return getAccountRole(store, userbase.getAccountNames(store))


def getAccountRole(store, accountNames):
    """
    Retrieve the first Role in the given store which corresponds an account
    name in C{accountNames}.

    Note: the implementation currently ignores all of the values in
    C{accountNames} except for the first.

    @param accountNames: A C{list} of two-tuples of account local parts and
        domains.

    @raise ValueError: If C{accountNames} is empty.

    @rtype: L{Role}
    """
    for (localpart, domain) in accountNames:
        return getPrimaryRole(store, u'%s@%s' % (localpart, domain),
                              createIfNotFound=True)
    raise ValueError("Cannot get named role for unnamed account.")



def shareItem(sharedItem, toRole=None, toName=None, shareID=None,
              interfaces=ALL_IMPLEMENTED):
    """
    Share an item with a given role.  This provides a way to expose items to
    users for later retrieval with L{Role.getShare}.

    This API is slated for deprecation.  Prefer L{Role.shareItem} in new code.

    @param sharedItem: an item to be shared.

    @param toRole: a L{Role} instance which represents the group that has
    access to the given item.  May not be specified if toName is also
    specified.

    @param toName: a unicode string which uniquely identifies a L{Role} in the
    same store as the sharedItem.

    @param shareID: a unicode string.  If provided, specify the ID under which
    the shared item will be shared.

    @param interfaces: a list of Interface objects which specify the methods
    and attributes accessible to C{toRole} on C{sharedItem}.

    @return: a L{Share} which records the ability of the given role to access
    the given item.
    """
    warnings.warn("Use Role.shareItem() instead of sharing.shareItem().",
                  PendingDeprecationWarning,
                  stacklevel=2)
    if toRole is None:
        if toName is not None:
            toRole = getPrimaryRole(sharedItem.store, toName, True)
        else:
            toRole = getEveryoneRole(sharedItem.store)
    return toRole.shareItem(sharedItem, shareID, interfaces)



def _linearize(interface):
    """
    Return a list of all the bases of a given interface in depth-first order.

    @param interface: an Interface object.

    @return: a L{list} of Interface objects, the input in all its bases, in
    subclass-to-base-class, depth-first order.
    """
    L = [interface]
    for baseInterface in interface.__bases__:
        if baseInterface is not Interface:
            L.extend(_linearize(baseInterface))
    return L



def _commonParent(zi1, zi2):
    """
    Locate the common parent of two Interface objects.

    @param zi1: a zope Interface object.

    @param zi2: another Interface object.

    @return: the rightmost common parent of the two provided Interface objects,
    or None, if they have no common parent other than Interface itself.
    """
    shorter, longer = sorted([_linearize(x)[::-1] for x in zi1, zi2],
                             key=len)
    for n in range(len(shorter)):
        if shorter[n] != longer[n]:
            if n == 0:
                return None
            return shorter[n-1]
    return shorter[-1]


def _checkConflictingNames(interfaces):
    """
    Raise an exception if any of the names present in the given interfaces
    conflict with each other.

    @param interfaces: a list of Zope Interface objects.

    @return: None

    @raise ConflictingNames: if any of the attributes of the provided
    interfaces are the same, and they do not have a common base interface which
    provides that name.
    """
    names = {}
    for interface in interfaces:
        for name in interface:
            if name in names:
                otherInterface = names[name]
                parent = _commonParent(interface, otherInterface)
                if parent is None or name not in parent:
                    raise ConflictingNames("%s conflicts with %s over %s" % (
                            interface, otherInterface, name))
            names[name] = interface



def getShare(store, role, shareID):
    """
    Retrieve the accessible facet of an Item previously shared with
    L{shareItem}.

    This method is pending deprecation, and L{Role.getShare} should be
    preferred in new code.

    @param store: an axiom store (XXX must be the same as role.store)

    @param role: a L{Role}, the primary role for a user attempting to retrieve
    the given item.

    @return: a L{SharedProxy}.  This is a wrapper around the shared item which
    only exposes those interfaces explicitly allowed for the given role.

    @raise: L{NoSuchShare} if there is no item shared to the given role for the
    given shareID.
    """
    warnings.warn("Use Role.getShare() instead of sharing.getShare().",
                  PendingDeprecationWarning,
                  stacklevel=2)
    return role.getShare(shareID)



def asAccessibleTo(role, query):
    """
    Return an iterable which yields the shared proxies that are available to
    the given role, from the given query.

    This method is pending deprecation, and L{Role.asAccessibleTo} should be
    preferred in new code.

    @param role: The role to retrieve L{SharedProxy}s for.

    @param query: An Axiom query describing the Items to retrieve, which this
    role can access.
    @type query: an L{iaxiom.IQuery} provider.
    """
    warnings.warn(
        "Use Role.asAccessibleTo() instead of sharing.asAccessibleTo().",
        PendingDeprecationWarning,
        stacklevel=2)
    return role.asAccessibleTo(query)



def itemFromProxy(obj):
    """
    Retrieve the real, underlying Item based on a L{SharedProxy} object, so
    that you can access all of its attributes and methods.

    This function is provided because sometimes it's hard to figure out how to
    cleanly achieve some behavior, especially running a query which relates to
    a shared proxy which you have retrieved.  However, if you find yourself
    calling it a lot, that's a very bad sign: calling this method is implicitly
    a breach of the security that the sharing system tries to provide.
    Normally, if your code is acting as an agent of role X, it has access to a
    L{SharedProxy} that only provides interfaces explicitly allowed to X.  If
    you make a mistake and call a method that the user is not supposed to be
    able to access, the user will receive an exception rather than be allowed
    to violate the system's security constraints.

    However, once you have retrieved the underlying item, all bets are off, and
    you have to perform your own security checks.  This is error-prone, and
    should be avoided.  We suggest, instead, adding explicitly allowed methods
    for performing any queries which your objects need.

    @param obj: a L{SharedProxy} instance

    @return: the underlying Item instance of the given L{SharedProxy}, with all
    of its methods and attributes exposed.
    """
    return object.__getattribute__(obj, '_sharedItem')



def unShare(sharedItem):
    """
    Remove all instances of this item from public or shared view.
    """
    sharedItem.store.query(Share, Share.sharedItem == sharedItem).deleteFromStore()



def randomEarlyShared(store, role):
    """
    If there are no explicitly-published public index pages to display, find a
    shared item to present to the user as first.
    """
    for r in role.allRoles():
        share = store.findFirst(Share, Share.sharedTo == r,
                                sort=Share.storeID.ascending)
        if share is not None:
            return share.sharedItem
    raise NoSuchShare("Why, that user hasn't shared anything at all!")




from axiom.test.historic import stubloader
from xmantissa.signup import FreeTicketSignup

class FreeTicketSignupTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        fts = self.store.findUnique(FreeTicketSignup)
        ae = self.assertEqual

        ae(fts.prefixURL, '/a/b')
        ae(fts.booth, self.store)
        ae(fts.emailTemplate, 'TEMPLATE!')
        ae(fts.prompt, 'Sign Up')



from axiom.test.historic import stubloader
from xmantissa.people import PhoneNumber, Person

class PhoneNumberTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        phone = self.store.findUnique(PhoneNumber)
        person = self.store.findUnique(Person)
        self.assertIdentical(phone.person, person)
        self.assertEquals(phone.number, '555-1212')



"""
Tests for xmantissa.offering.
"""

from zope.interface import Interface, implements
from zope.interface.verify import verifyClass, verifyObject

from twisted.trial import unittest

from axiom.store import Store
from axiom import item, attributes, userbase

from axiom.plugins.mantissacmd import Mantissa

from axiom.dependency import installedOn

from xmantissa import ixmantissa, offering

from xmantissa.web import SiteConfiguration
from xmantissa.ampserver import AMPConfiguration
from xmantissa.plugins.baseoff import baseOffering, ampOffering
from xmantissa.plugins.offerings import peopleOffering


class TestSiteRequirement(item.Item):
    typeName = 'test_site_requirement'
    schemaVersion = 1

    attr = attributes.integer()

class TestAppPowerup(item.Item):
    typeName = 'test_app_powerup'
    schemaVersion = 1

    attr = attributes.integer()



class ITestInterface(Interface):
    """
    An interface to which no object can be adapted.  Used to ensure failed
    adaption causes a powerup to be installed.
    """



class OfferingPluginTest(unittest.TestCase):
    """
    A simple test for getOffering.
    """

    def test_getOfferings(self):
        """
        getOffering should use the Twisted plugin system to load the plugins
        provided with Mantissa.  Since this is dynamic, we can't assert
        anything about the complete list, but we can at least verify that all
        the plugins that should be there, are.
        """
        foundOfferings = list(offering.getOfferings())
        allExpectedOfferings = [baseOffering, ampOffering, peopleOffering]
        for expected in allExpectedOfferings:
            self.assertIn(expected, foundOfferings)


class OfferingTest(unittest.TestCase):
    def setUp(self):
        self.store = Store(filesdir=self.mktemp())
        Mantissa().installSite(self.store, u"localhost", u"", False)
        Mantissa().installAdmin(self.store, u'admin', u'localhost', u'asdf')
        self.userbase = self.store.findUnique(userbase.LoginSystem)
        self.adminAccount = self.userbase.accountByAddress(
            u'admin', u'localhost')
        off = offering.Offering(
            name=u'test_offering',
            description=u'This is an offering which tests the offering '
                         'installation mechanism',
            siteRequirements=[(ITestInterface, TestSiteRequirement)],
            appPowerups=[TestAppPowerup],
            installablePowerups=[],
            loginInterfaces=[],
            themes=[],
            )
        self.offering = off
        # Add this somewhere that the plugin system is going to see it.
        self._originalGetOfferings = offering.getOfferings
        offering.getOfferings = self.fakeGetOfferings


    def fakeGetOfferings(self):
        """
        Return standard list of offerings, plus one extra.
        """
        return list(self._originalGetOfferings()) + [self.offering]


    def tearDown(self):
        """
        Remove the temporary offering.
        """
        offering.getOfferings = self._originalGetOfferings


    def test_installOffering(self):
        """
        L{OfferingConfiguration.installOffering} should install the given
        offering on the Mantissa server.
        """
        conf = self.adminAccount.avatars.open().findUnique(
            offering.OfferingConfiguration)
        io = conf.installOffering(self.offering, None)

        # InstalledOffering should be returned, and installed on the site store
        foundIO = self.store.findUnique(offering.InstalledOffering,
                  offering.InstalledOffering.offeringName == self.offering.name)
        self.assertIdentical(io, foundIO)

        # Site store requirements should be on the site store
        tsr = self.store.findUnique(TestSiteRequirement)
        self.failUnless(installedOn(tsr), self.store)

        # App store should have been created
        appStore = self.userbase.accountByAddress(self.offering.name, None)
        self.assertNotEqual(appStore, None)

        # App store requirements should be on the app store
        ss = appStore.avatars.open()
        tap = ss.findUnique(TestAppPowerup)
        self.failUnless(installedOn(tap), ss)

        self.assertRaises(offering.OfferingAlreadyInstalled,
                          conf.installOffering, self.offering, None)


    def test_getInstalledOfferingNames(self):
        """
        L{getInstalledOfferingNames} should list the names of offerings
        installed on the given site store.
        """
        self.assertEquals(offering.getInstalledOfferingNames(self.store),
                          ['mantissa-base'])

        self.test_installOffering()

        installed = offering.getInstalledOfferingNames(self.store)
        installed.sort()
        expected = [u"mantissa-base", u"test_offering"]
        expected.sort()
        self.assertEquals(installed, expected)


    def test_getInstalledOfferings(self):
        """
        getInstalledOfferings should return a mapping of offering name to
        L{Offering} object for each installed offering on a given site store.
        """
        self.assertEquals(offering.getInstalledOfferings(self.store),
                          {baseOffering.name: baseOffering})
        self.test_installOffering()
        self.assertEquals(offering.getInstalledOfferings(self.store),
                          {baseOffering.name: baseOffering,
                           self.offering.name: self.offering})


    def test_isAppStore(self):
        """
        isAppStore returns True for stores with offerings installed on them,
        False otherwise.
        """
        conf = self.adminAccount.avatars.open().findUnique(
            offering.OfferingConfiguration)
        conf.installOffering(self.offering, None)
        app = self.userbase.accountByAddress(self.offering.name, None)
        self.failUnless(offering.isAppStore(app.avatars.open()))
        self.failIf(offering.isAppStore(self.adminAccount.avatars.open()))



class FakeOfferingTechnician(object):
    """
    In-memory only implementation of the offering inspection/installation API.

    @ivar installedOfferings: A mapping from offering names to corresponding
        L{IOffering} providers which have been passed to C{installOffering}.
    """
    implements(ixmantissa.IOfferingTechnician)

    def __init__(self):
        self.installedOfferings = {}


    def installOffering(self, offering):
        """
        Add the given L{IOffering} provider to the list of installed offerings.
        """
        self.installedOfferings[offering.name] = offering


    def getInstalledOfferings(self):
        """
        Return a copy of the internal installed offerings mapping.
        """
        return self.installedOfferings.copy()


    def getInstalledOfferingNames(self):
        """
        Return the names from the internal installed offerings mapping.
        """
        return self.installedOfferings.keys()



class OfferingTechnicianTestMixin:
    """
    L{unittest.TestCase} mixin which defines unit tests for classes which
    implement L{IOfferingTechnician}.

    @ivar offerings: A C{list} of L{Offering} instances which will be installed
        by the tests this mixin defines.
    """
    offerings = [
        offering.Offering(u'an offering', None, [], [], [], [], []),
        offering.Offering(u'another offering', None, [], [], [], [], [])]

    def createTechnician(self):
        """
        @return: An L{IOfferingTechnician} provider which will be tested.
        """
        raise NotImplementedError(
            "%r did not implement createTechnician" % (self.__class__,))


    def test_interface(self):
        """
        L{createTechnician} returns an instance of a type which declares that
        it implements L{IOfferingTechnician} and has all of the methods and
        attributes defined by the interface.
        """
        technician = self.createTechnician()
        technicianType = type(technician)
        self.assertTrue(
            ixmantissa.IOfferingTechnician.implementedBy(technicianType))
        self.assertTrue(
            verifyClass(ixmantissa.IOfferingTechnician, technicianType))
        self.assertTrue(
            verifyObject(ixmantissa.IOfferingTechnician, technician))


    def test_getInstalledOfferingNames(self):
        """
        The L{ixmantissa.IOfferingTechnician.getInstalledOfferingNames}
        implementation returns a C{list} of C{unicode} strings, each element
        giving the name of an offering which has been installed.
        """
        offer = self.createTechnician()
        self.assertEqual(offer.getInstalledOfferingNames(), [])

        expected = []
        for dummyOffering in self.offerings:
            offer.installOffering(dummyOffering)
            expected.append(dummyOffering.name)
            expected.sort()
            installed = offer.getInstalledOfferingNames()
            installed.sort()
            self.assertEqual(installed, expected)


    def test_getInstalledOfferings(self):
        """
        The L{ixmantissa.IOfferingTechnician.getInstalledOfferings}
        implementation returns a C{dict} mapping C{unicode} offering names to
        the corresponding L{IOffering} providers.
        """
        offer = self.createTechnician()
        self.assertEqual(offer.getInstalledOfferings(), {})

        expected = {}
        for dummyOffering in self.offerings:
            offer.installOffering(dummyOffering)
            expected[dummyOffering.name] = dummyOffering
            self.assertEqual(offer.getInstalledOfferings(), expected)



class OfferingAdapterTests(unittest.TestCase, OfferingTechnicianTestMixin):
    """
    Tests for L{offering.OfferingAdapter}.
    """
    def setUp(self):
        """
        Hook offering plugin discovery so that only the fake offerings the test
        wants exist.
        """
        self.origGetOfferings = offering.getOfferings
        offering.getOfferings = self.getOfferings


    def tearDown(self):
        """
        Restore the original L{getOfferings} function.
        """
        offering.getOfferings = self.origGetOfferings


    def getOfferings(self):
        """
        Return some dummy offerings, as defined by C{self.offerings}.
        """
        return self.offerings


    def createTechnician(self):
        """
        Create an L{offering.OfferingAdapter}.
        """
        store = Store()
        technician = offering.OfferingAdapter(store)
        return technician



class FakeOfferingTechnicianTests(unittest.TestCase, OfferingTechnicianTestMixin):
    """
    Tests (ie, verification) for L{FakeOfferingTechnician}.
    """
    def createTechnician(self):
        """
        Create a L{FakeOfferingTechnician}.
        """
        return FakeOfferingTechnician()



class BaseOfferingTests(unittest.TestCase):
    """
    Tests for the base Mantissa offering,
    L{xmantissa.plugins.baseoff.baseOffering}.
    """
    def test_interface(self):
        """
        C{baseOffering} provides L{IOffering}.
        """
        self.assertTrue(verifyObject(ixmantissa.IOffering, baseOffering))


    def test_staticContentPath(self):
        """
        C{baseOffering.staticContentPath} gives the location of a directory
        which has I{mantissa.css} in it.
        """
        self.assertTrue(
            baseOffering.staticContentPath.child('mantissa.css').exists())


    def _siteRequirementTest(self, offering, cls):
        """
        Verify that installing C{offering} results in an instance of the given
        Item subclass being installed as a powerup for IProtocolFactoryFactory.
        """
        store = Store()
        ixmantissa.IOfferingTechnician(store).installOffering(offering)

        factories = list(store.powerupsFor(ixmantissa.IProtocolFactoryFactory))
        for factory in factories:
            if isinstance(factory, cls):
                break
        else:
            self.fail("No instance of %r in %r" % (cls, factories))


    def test_siteConfiguration(self):
        """
        L{SiteConfiguration} powers up a store for L{IProtocolFactoryFactory}
        when L{baseOffering} is installed on that store.
        """
        self._siteRequirementTest(baseOffering, SiteConfiguration)


    def test_ampConfiguration(self):
        """
        L{AMPConfiguration} powers up a store for L{IProtocolFactoryFactory}
        when L{ampOffering} is installed on that store.
        """
        self._siteRequirementTest(ampOffering, AMPConfiguration)



# -*- test-case-name: xmantissa.test.test_webapp -*-

"""
This module is the basis for Mantissa-based web applications.  It provides
several basic pluggable application-level features, most notably Powerup-based
integration of the extensible hierarchical navigation system defined in
xmantissa.webnav
"""

import os

from zope.interface import implements

from epsilon.structlike import record

from axiom.iaxiom import IPowerupIndirector
from axiom.item import Item, declareLegacyItem
from axiom.attributes import text, integer, reference
from axiom import upgrade
from axiom.dependency import dependsOn
from axiom.userbase import getAccountNames

from nevow.rend import Page
from nevow import livepage, athena
from nevow.inevow import IRequest
from nevow import tags as t
from nevow import url

from xmantissa.publicweb import CustomizedPublicPage, renderShortUsername

from xmantissa.ixmantissa import (
    INavigableElement, ISiteRootPlugin, IWebTranslator, IStaticShellContent,
    ITemplateNameResolver, ISiteURLGenerator, IWebViewer)

from xmantissa.website import PrefixURLMixin, JUST_SLASH, WebSite, APIKey
from xmantissa.website import MantissaLivePage
from xmantissa.webtheme import getInstalledThemes
from xmantissa.webnav import getTabs, startMenu, settingsLink, applicationNavigation
from xmantissa.sharing import getPrimaryRole

from xmantissa._webidgen import genkey, storeIDToWebID, webIDToStoreID
from xmantissa._webutil import MantissaViewHelper, WebViewerHelper
from xmantissa.offering import getInstalledOfferings

from xmantissa.webgestalt import AuthenticationApplication
from xmantissa.prefs import PreferenceAggregator, DefaultPreferenceCollection
from xmantissa.search import SearchAggregator


def _reorderForPreference(themeList, preferredThemeName):
    """
    Re-order the input themeList according to the preferred theme.

    Returns None.
    """
    for theme in themeList:
        if preferredThemeName == theme.themeName:
            themeList.remove(theme)
            themeList.insert(0, theme)
            return

class _WebIDFormatException(TypeError):
    """
    An inbound web ID was not formatted as expected.
    """



class _AuthenticatedWebViewer(WebViewerHelper):
    """
    Implementation of L{IWebViewer} for authenticated users.

    @ivar _privateApplication: the L{PrivateApplication} for the authenticated
    user that this view is rendering.
    """
    implements(IWebViewer)

    def __init__(self, privateApp):
        """
        @param privateApp: Probably something abstract but really it's just a
        L{PrivateApplication}.
        """
        WebViewerHelper.__init__(
            self, privateApp.getDocFactory, privateApp._preferredThemes)
        self._privateApplication = privateApp


    # IWebViewer
    def roleIn(self, userStore):
        """
        Get the authenticated role for the user represented by this view in the
        given user store.
        """
        return getPrimaryRole(userStore, self._privateApplication._getUsername())


    # Complete WebViewerHelper implementation
    def _wrapNavFrag(self, frag, useAthena):
        """
        Wrap the given L{INavigableFragment} in an appropriate
        L{_FragmentWrapperMixin} subclass.
        """
        username = self._privateApplication._getUsername()
        cf = getattr(frag, 'customizeFor', None)
        if cf is not None:
            frag = cf(username)
        if useAthena:
            pageClass = GenericNavigationAthenaPage
        else:
            pageClass = GenericNavigationPage
        return pageClass(self._privateApplication, frag,
                         self._privateApplication.getPageComponents(),
                         username)



class _ShellRenderingMixin(object):
    """
    View mixin for Pages which use the I{shell} template.

    This class provides somewhat sensible default implementations for a number
    of the renderers required by the I{shell} template.

    @ivar webapp: The L{PrivateApplication} of the user for whom this is a
        view.  This must provide the I{rootURL} method as well as
        L{IWebTranslator} and L{ITemplateNameResolver}.  This must be an item
        in a user store associated with the site store (so that the site store
        is available).
    """
    fragmentName = 'main'
    searchPattern = None

    def __init__(self, webapp, pageComponents, username):
        self.webapp = self.translator = self.resolver = webapp
        self.pageComponents = pageComponents
        self.username = username


    def _siteStore(self):
        """
        Get the site store from C{self.webapp}.
        """
        return self.webapp.store.parent


    def getDocFactory(self, fragmentName, default=None):
        """
        Retrieve a Nevow document factory for the given name. This
        implementation merely defers to the underlying L{PrivateApplication}.

        @param fragmentName: a short string that names a fragment template.

        @param default: value to be returned if the named template is not
        found.
        """
        return self.webapp.getDocFactory(fragmentName, default)


    def render_content(self, ctx, data):
        raise NotImplementedError("implement render_context in subclasses")


    def render_title(self, ctx, data):
        return ctx.tag[self.__class__.__name__]


    def render_rootURL(self, ctx, data):
        """
        Add the WebSite's root URL as a child of the given tag.

        The root URL is the location of the resource beneath which all standard
        Mantissa resources (such as the private application and static content)
        is available.  This can be important if a page is to be served at a
        location which is different from the root URL in order to make links in
        static XHTML templates resolve correctly (for example, by adding this
        value as the href of a <base> tag).
        """
        site = ISiteURLGenerator(self._siteStore())
        return ctx.tag[site.rootURL(IRequest(ctx))]


    def render_head(self, ctx, data):
        return ctx.tag

    def render_header(self, ctx, data):
        staticShellContent = self.pageComponents.staticShellContent
        if staticShellContent is None:
            return ctx.tag
        header = staticShellContent.getHeader()
        if header is not None:
            return ctx.tag[header]
        else:
            return ctx.tag


    def render_startmenu(self, ctx, data):
        """
        Add start-menu style navigation to the given tag.

        @see {xmantissa.webnav.startMenu}
        """
        return startMenu(
            self.translator, self.pageComponents.navigation, ctx.tag)


    def render_settingsLink(self, ctx, data):
        """
        Add the URL of the settings page to the given tag.

        @see L{xmantissa.webnav.settingsLink}
        """
        return settingsLink(
            self.translator, self.pageComponents.settings, ctx.tag)


    def render_applicationNavigation(self, ctx, data):
        """
        Add primary application navigation to the given tag.

        @see L{xmantissa.webnav.applicationNavigation}
        """
        return applicationNavigation(
            ctx, self.translator, self.pageComponents.navigation)


    def render_urchin(self, ctx, data):
        """
        Render the code for recording Google Analytics statistics, if so
        configured.
        """
        key = APIKey.getKeyForAPI(self._siteStore(), APIKey.URCHIN)
        if key is None:
            return ''
        return ctx.tag.fillSlots('urchin-key', key.apiKey)


    def render_search(self, ctx, data):
        searchAggregator = self.pageComponents.searchAggregator
        if searchAggregator is None or not searchAggregator.providers():
            return ''
        return ctx.tag.fillSlots(
            'form-action', self.translator.linkTo(searchAggregator.storeID))


    def render_username(self, ctx, data):
        return renderShortUsername(ctx, self.username)


    def render_logout(self, ctx, data):
        return ctx.tag


    def render_authenticateLinks(self, ctx, data):
        return ''


    def _getVersions(self):
        versions = []
        for (name, offering) in getInstalledOfferings(self._siteStore()).iteritems():
            if offering.version is not None:
                v = offering.version
                versions.append(str(v).replace(v.package, name))
        return ' '.join(versions)


    def render_footer(self, ctx, data):
        footer = [self._getVersions()]
        staticShellContent = self.pageComponents.staticShellContent
        if staticShellContent is not None:
            f = staticShellContent.getFooter()
            if f is not None:
                footer.append(f)
        return ctx.tag[footer]



INSPECTROFY = os.environ.get('MANTISSA_DEV')

class _FragmentWrapperMixin(MantissaViewHelper):
    def __init__(self, fragment, pageComponents):
        self.fragment = fragment
        fragment.page = self
        self.pageComponents = pageComponents

    def beforeRender(self, ctx):
        return getattr(self.fragment, 'beforeRender', lambda x: None)(ctx)

    def render_introspectionWidget(self, ctx, data):
        "Until we have eliminated everything but GenericAthenaLivePage"
        if INSPECTROFY:
            return ctx.tag['No debugging on crap-ass bad pages']
        else:
            return ''

    def render_head(self, ctx, data):
        req = IRequest(ctx)

        userStore = self.webapp.store
        siteStore = userStore.parent
        site = ISiteURLGenerator(siteStore)

        l = self.pageComponents.themes
        _reorderForPreference(l, self.webapp.preferredTheme)
        extras = []
        for theme in l:
            extra = theme.head(req, site)
            if extra is not None:
                extras.append(extra)
        headMethod = getattr(self.fragment, 'head', None)
        if headMethod is not None:
            extra = headMethod()
            if extra is not None:
                extras.append(extra)
        return ctx.tag[extras]


    def render_title(self, ctx, data):
        """
        Return the current context tag containing C{self.fragment}'s C{title}
        attribute, or "Divmod".
        """
        return ctx.tag[getattr(self.fragment, 'title', 'Divmod')]

    def render_content(self, ctx, data):
        return ctx.tag[self.fragment]

class GenericNavigationPage(_FragmentWrapperMixin, Page, _ShellRenderingMixin):
    def __init__(self, webapp, fragment, pageComponents, username):
        Page.__init__(self, docFactory=webapp.getDocFactory('shell'))
        _ShellRenderingMixin.__init__(self, webapp, pageComponents, username)
        _FragmentWrapperMixin.__init__(self, fragment, pageComponents)


class GenericNavigationLivePage(_FragmentWrapperMixin, livepage.LivePage, _ShellRenderingMixin):
    def __init__(self, webapp, fragment, pageComponents, username):
        livepage.LivePage.__init__(self, docFactory=webapp.getDocFactory('shell'))
        _ShellRenderingMixin.__init__(self, webapp, pageComponents, username)
        _FragmentWrapperMixin.__init__(self, fragment, pageComponents)

    # XXX TODO: support live nav, live fragments somehow
    def render_head(self, ctx, data):
        ctx.tag[t.invisible(render=t.directive("liveglue"))]
        return _FragmentWrapperMixin.render_head(self, ctx, data)

    def goingLive(self, ctx, client):
        getattr(self.fragment, 'goingLive', lambda x, y: None)(ctx, client)

    def locateHandler(self, ctx, path, name):
        handler = getattr(self.fragment, 'locateHandler', None)

        if handler is None:
            return getattr(self.fragment, 'handle_' + name)
        else:
            return handler(ctx, path, name)



class GenericNavigationAthenaPage(_FragmentWrapperMixin,
                                  MantissaLivePage,
                                  _ShellRenderingMixin):
    """
    This class provides the generic navigation elements for surrounding all
    pages navigated under the /private/ namespace.
    """
    def __init__(self, webapp, fragment, pageComponents, username):
        """
        Top-level container for Mantissa application views.

        @param webapp: a C{PrivateApplication}.
        @param fragment: The C{Element} or C{Fragment} to display as content.
        @param pageComponents a C{_PageComponent}.

        This page draws its HTML from the 'shell' template in the preferred
        theme for the store.  If loaded in a browser that does not support
        Athena, the page provided by the 'athena-unsupported' template will be
        displayed instead.

        @see: L{PrivateApplication.preferredTheme}
        """
        userStore = webapp.store
        siteStore = userStore.parent

        MantissaLivePage.__init__(
            self, ISiteURLGenerator(siteStore),
            getattr(fragment, 'iface', None),
            fragment,
            jsModuleRoot=None,
            docFactory=webapp.getDocFactory('shell'))
        _ShellRenderingMixin.__init__(self, webapp, pageComponents, username)
        _FragmentWrapperMixin.__init__(self, fragment, pageComponents)
        self.unsupportedBrowserLoader = (webapp
                                         .getDocFactory("athena-unsupported"))


    def beforeRender(self, ctx):
        """
        Call the C{beforeRender} implementations on L{MantissaLivePage} and
        L{_FragmentWrapperMixin}.
        """
        MantissaLivePage.beforeRender(self, ctx)
        return _FragmentWrapperMixin.beforeRender(self, ctx)


    def render_head(self, ctx, data):
        ctx.tag[t.invisible(render=t.directive("liveglue"))]
        return _FragmentWrapperMixin.render_head(self, ctx, data)


    def render_introspectionWidget(self, ctx, data):
        if INSPECTROFY:
            f = athena.IntrospectionFragment()
            f.setFragmentParent(self)
            return ctx.tag[f]
        else:
            return ''



class _PrivateRootPage(Page, _ShellRenderingMixin):
    """
    L{_PrivateRootPage} is the resource present for logged-in users at
    "/private", providing a direct interface to the objects located in the
    user's personal user-store.

    It is created by L{PrivateApplication.createResourceWith}.
    """
    addSlash = True

    def __init__(self, webapp, pageComponents, username, webViewer):
        self.username = username
        self.webViewer = webViewer
        Page.__init__(self, docFactory=webapp.getDocFactory('shell'))
        _ShellRenderingMixin.__init__(self, webapp, pageComponents, username)

    def child_(self, ctx):
        navigation = self.pageComponents.navigation
        if not navigation:
            return self
        # /private/XXXX ->
        click = self.webapp.linkTo(navigation[0].storeID)
        return url.URL.fromContext(ctx).click(click)

    def render_content(self, ctx, data):
        return """
        You have no default root page set, and no navigation plugins installed.  I
        don't know what to do.
        """


    def render_title(self, ctx, data):
        return ctx.tag['Private Root Page (You Should Not See This)']


    def childFactory(self, ctx, name):
        """
        Return a shell page wrapped around the Item model described by the
        webID, or return None if no such item can be found.
        """
        try:
            o = self.webapp.fromWebID(name)
        except _WebIDFormatException:
            return None
        if o is None:
            return None
        return self.webViewer.wrapModel(o)



class _PageComponents(record('navigation searchAggregator staticShellContent settings themes')):
    """
    I encapsulate various plugin objects that have some say
    in determining the available functionality on a given page
    """
    pass

class PrivateApplication(Item, PrefixURLMixin):
    """
    This is the root of a private, navigable web application.  It is designed
    to be installed on avatar stores after installing WebSite.

    To plug into it, install powerups of the type INavigableElement on the
    user's store.  Their tabs will be retrieved and items that are part of
    those powerups will be linked to; provide adapters for said items to either
    INavigableFragment or IResource.

    Note: IResource adapters should be used sparingly, for example, for
    specialized web resources which are not 'nodes' within the application; for
    example, that need to set a custom content/type or that should not display
    any navigation elements because they will be displayed only within IFRAME
    nodes.  Do _NOT_ use IResource adapters to provide a customized
    look-and-feel; instead use mantissa themes.  (XXX document webtheme.py more
    thoroughly)

    @ivar preferredTheme: A C{unicode} string naming the preferred theme for
    this application.  Templates and suchlike will be looked up for this theme
    first.

    @ivar privateKey: A random integer used to deterministically but
    unpredictably perturb link generation to avoid being the target of XSS
    attacks.

    @ivar privateIndexPage: A reference to the Item whose IResource or
    INavigableFragment adapter will be displayed on login and upon viewing the
    'root' page, /private/.
    """

    implements(ISiteRootPlugin, IWebTranslator, ITemplateNameResolver,
               IPowerupIndirector)

    powerupInterfaces = (IWebTranslator, ITemplateNameResolver, IWebViewer)

    typeName = 'private_web_application'
    schemaVersion = 5

    preferredTheme = text()
    privateKey = integer(defaultFactory=genkey)

    website = dependsOn(WebSite)

    customizedPublicPage = dependsOn(CustomizedPublicPage)
    authenticationApplication = dependsOn(AuthenticationApplication)
    preferenceAggregator = dependsOn(PreferenceAggregator)
    defaultPreferenceCollection = dependsOn(DefaultPreferenceCollection)
    searchAggregator = dependsOn(SearchAggregator)

    #XXX Nothing ever uses this
    privateIndexPage = reference()

    prefixURL = 'private'

    sessioned = True
    sessionless = False

    def getPageComponents(self):
        navigation = getTabs(self.store.powerupsFor(INavigableElement))

        staticShellContent = IStaticShellContent(self.store, None)

        return _PageComponents(navigation,
                               self.searchAggregator,
                               staticShellContent,
                               self.store.findFirst(PreferenceAggregator),
                               getInstalledThemes(self.store.parent))


    def _getUsername(self):
        """
        Return a localpart@domain style string naming the owner of our store.

        @rtype: C{unicode}
        """
        for (l, d) in getAccountNames(self.store):
            return l + u'@' + d


    def createResourceWith(self, webViewer):
        return _PrivateRootPage(self, self.getPageComponents(),
                                self._getUsername(), webViewer)


    # ISiteRootPlugin
    def produceResource(self, req, segments, webViewer):
        if segments == JUST_SLASH:
            return self.createResourceWith(webViewer), JUST_SLASH
        else:
            return super(PrivateApplication, self).produceResource(
                req, segments, webViewer)


    # IWebTranslator
    def linkTo(self, obj):
        # currently obj must be a storeID, but other types might come eventually
        return '/%s/%s' % (self.prefixURL, storeIDToWebID(self.privateKey, obj))

    def linkToWithActiveTab(self, childItem, parentItem):
        """
        Return a URL which will point to the web facet of C{childItem},
        with the selected nav tab being the one that represents C{parentItem}
        """
        return self.linkTo(parentItem.storeID) + '/' + self.toWebID(childItem)

    def linkFrom(self, webid):
        return webIDToStoreID(self.privateKey, webid)

    def fromWebID(self, webID):
        storeID = self.linkFrom(webID)
        if storeID is None:
            # This is not a very good interface, but I don't want to change the
            # calling code right now as I'm neither confident in its test
            # coverage nor looking to go on a test-writing rampage through this
            # code for a minor fix.
            raise _WebIDFormatException("%r didn't look like a webID" % (webID,))
        webitem = self.store.getItemByID(storeID, None)
        return webitem

    def toWebID(self, item):
        return storeIDToWebID(self.privateKey, item.storeID)


    def _preferredThemes(self):
        """
        Return a list of themes in the order of preference that this user has
        selected via L{PrivateApplication.preferredTheme}.
        """
        themes = getInstalledThemes(self.store.parent)
        _reorderForPreference(themes, self.preferredTheme)
        return themes


    #ITemplateNameResolver
    def getDocFactory(self, fragmentName, default=None):
        """
        Retrieve a Nevow document factory for the given name.

        @param fragmentName: a short string that names a fragment template.

        @param default: value to be returned if the named template is not
        found.
        """
        themes = self._preferredThemes()
        for t in themes:
            fact = t.getDocFactory(fragmentName, None)
            if fact is not None:
                return fact
        return default


    # IPowerupIndirector
    def indirect(self, interface):
        """
        Indirect the implementation of L{IWebViewer} to
        L{_AuthenticatedWebViewer}.
        """
        if interface == IWebViewer:
            return _AuthenticatedWebViewer(self)
        return self



PrivateApplicationV2 = declareLegacyItem(PrivateApplication.typeName, 2, dict(
    installedOn = reference(),
    preferredTheme = text(),
    hitCount = integer(default=0),
    privateKey = integer(),
    privateIndexPage = reference(),
    ))

PrivateApplicationV3 = declareLegacyItem(PrivateApplication.typeName, 3, dict(
    preferredTheme=text(),
    hitCount=integer(default=0),
    privateKey=integer(),
    privateIndexPage=reference(),
    customizedPublicPage=reference("dependsOn(CustomizedPublicPage)"),
    authenticationApplication=reference("dependsOn(AuthenticationApplication)"),
    preferenceAggregator=reference("dependsOn(PreferenceAggregator)"),
    defaultPreferenceCollection=reference("dependsOn(DefaultPreferenceCollection)"),
    searchAggregator=reference("dependsOn(SearchAggregator)"),
    website=reference(),
    ))

def upgradePrivateApplication1To2(oldApp):
    newApp = oldApp.upgradeVersion(
        'private_web_application', 1, 2,
        installedOn=oldApp.installedOn,
        preferredTheme=oldApp.preferredTheme,
        privateKey=oldApp.privateKey,
        privateIndexPage=oldApp.privateIndexPage)
    newApp.store.powerup(newApp.store.findOrCreate(
        CustomizedPublicPage), ISiteRootPlugin, -257)
    return newApp

upgrade.registerUpgrader(upgradePrivateApplication1To2, 'private_web_application', 1, 2)

def _upgradePrivateApplication2to3(old):
    pa = old.upgradeVersion(PrivateApplication.typeName, 2, 3,
        preferredTheme=old.preferredTheme,
        privateKey=old.privateKey,
        privateIndexPage=old.privateIndexPage)
    pa.customizedPublicPage = old.store.findOrCreate(CustomizedPublicPage)
    pa.authenticationApplication = old.store.findOrCreate(AuthenticationApplication)
    pa.preferenceAggregator = old.store.findOrCreate(PreferenceAggregator)
    pa.defaultPreferenceCollection = old.store.findOrCreate(DefaultPreferenceCollection)
    pa.searchAggregator = old.store.findOrCreate(SearchAggregator)
    pa.website = old.store.findOrCreate(WebSite)
    return pa

upgrade.registerUpgrader(_upgradePrivateApplication2to3, PrivateApplication.typeName, 2, 3)

def upgradePrivateApplication3to4(old):
    """
    Upgrade L{PrivateApplication} from schema version 3 to schema version 4.

    Copy all existing attributes to the new version and use the
    L{PrivateApplication} to power up the item it is installed on for
    L{ITemplateNameResolver}.
    """
    new = old.upgradeVersion(
        PrivateApplication.typeName, 3, 4,
        preferredTheme=old.preferredTheme,
        privateKey=old.privateKey,
        website=old.website,
        customizedPublicPage=old.customizedPublicPage,
        authenticationApplication=old.authenticationApplication,
        preferenceAggregator=old.preferenceAggregator,
        defaultPreferenceCollection=old.defaultPreferenceCollection,
        searchAggregator=old.searchAggregator)
    # Almost certainly this would be more correctly expressed as
    # installedOn(new).powerUp(...), however the 2 to 3 upgrader failed to
    # translate the installedOn attribute to state which installedOn can
    # recognize, consequently installedOn(new) will return None for an item
    # which was created at schema version 2 or earlier.  It's not worth dealing
    # with this inconsistency, since PrivateApplication is always only
    # installed on its store. -exarkun
    new.store.powerUp(new, ITemplateNameResolver)
    return new

upgrade.registerUpgrader(upgradePrivateApplication3to4, PrivateApplication.typeName, 3, 4)

PrivateApplicationV4 = declareLegacyItem(
    'private_web_application', 4,
    dict(authenticationApplication=reference(),
         customizedPublicPage=reference(),
         defaultPreferenceCollection=reference(),
         hitCount=integer(),
         preferenceAggregator=reference(),
         preferredTheme=text(),
         privateIndexPage=reference(),
         privateKey=integer(),
         searchAggregator=reference(),
         website=reference()))

def upgradePrivateApplication4to5(old):
    """
    Install the newly required powerup.
    """
    new = old.upgradeVersion(
        PrivateApplication.typeName, 4, 5,
        preferredTheme=old.preferredTheme,
        privateKey=old.privateKey,
        website=old.website,
        customizedPublicPage=old.customizedPublicPage,
        authenticationApplication=old.authenticationApplication,
        preferenceAggregator=old.preferenceAggregator,
        defaultPreferenceCollection=old.defaultPreferenceCollection,
        searchAggregator=old.searchAggregator)
    new.store.powerUp(new, IWebViewer)
    return new


upgrade.registerUpgrader(upgradePrivateApplication4to5, PrivateApplication.typeName, 4, 5)



from axiom.item import Item
from axiom import attributes
from axiom.upgrade import registerDeletionUpgrader

class Settings(Item):
    typeName = 'mantissa_settings'
    schemaVersion = 2

    installedOn = attributes.reference()

def settings1to2(old):
    new = old.upgradeVersion('mantissa_settings', 1, 2,
                             installedOn=None)
    new.deleteFromStore()

registerDeletionUpgrader(Settings, 1, 2)



# -*- test-case-name: xmantissa.test -*-
# Copyright 2008 Divmod, Inc. See LICENSE file for details
"""
Public interfaces used in Mantissa.
"""

from zope.interface import Interface, Attribute

from nevow.inevow import IRenderer


class IColumn(Interface):
    """
    Represents a column that can be viewed via a scrolling table, and provides
    hints & metadata about the column.
    """

    def sortAttribute():
        """
        @return: a sortable axiom.attribute, or None if this column cannot be
        sorted
        """


    def extractValue(model, item):
        """
        @type model: L{xmantissa.tdb.TabularDataModel}
        @param item: the L{axiom.item.Item} from which to extract column value

        @return: the underlying value for this column
        """


    def getType():
        """
        returns a string describing the type of this column, or None
        """


    def toComparableValue(value):
        """
        Convert a value received from the client into one that can be compared
        like-for-like with L{sortAttribute}, when executing an axiom query.

        (Callers should note that this is new as of Mantissa 0.6.6, and be
        prepared to deal with its absence in legacy code.)
        """


    attributeID = Attribute(
        """
        An ASCII-encoded str object uniquely describing this column.
        """)



class ITemplateNameResolver(Interface):
    """
    Loads Nevow document factories from a particular theme based on simple
    string names.
    """

    def getDocFactory(name, default=None):
        """
        Retrieve a Nevow document factory for the given name.

        @param name: a short string that names a fragment template for
        development purposes.

        @return: a Nevow docFactory
        """



class IPreferenceAggregator(Interface):
    """
    Allows convenient retrieval of individual preferences
    """

    def getPreferenceCollections():
        """
        Return a list of all installed L{IPreferenceCollection}s
        """

    def getPreferenceValue(key):
        """
        Return the value of the preference associated with "key"
        """

class ISearchProvider(Interface):
    """
    Represents an Item capable of searching for things
    """

    def count(term):
        """
        Return the number of items currently associated with the given
        (unprocessed) search string
        """

    def search(term, keywords=None, count=None, offset=0, sortAscending=True):
        """
        Query for items which contain the given term.

        @type term: C{unicode}
        @param keywords: C{dict} mapping C{unicode} field name to C{unicode}
        field contents.  Search results will be limited to documents with
        fields of these names containing these values.
        @type count: C{int} or C{NoneType}
        @type offset: C{int}, default is 0
        @param sortAscending: should the results be sorted ascendingly
        @type sortAscending: boolean

        @rtype: L{twisted.internet.defer.Deferred}
        @return: a Deferred which will fire with an iterable of
        L{search.SearchResult} instances, representing C{count} results for the
        unprocessed search represented by C{term}, starting at C{offset}.  The
        bounds of offset and count will be within the value last returned from
        L{count} for this term.
        """



class ISearchAggregator(Interface):
    """
    An Item responsible for interleaving and displaying search results
    obtained from available ISearchProviders
    """

    def count(term):
        """
        same as ISearchProvider.count, but queries all search providers
        """

    def search(term, keywords, count, offset, sortAscending):
        """
        same as ISearchProvider.search, but queries all search providers
        """

    def providers():
        """
        returns the number of available search providers
        """



class IFulltextIndexer(Interface):
    """
    A general interface to a low-level full-text indexer.
    """
    def add(document):
        """
        Add the given document to this index.

        This method may only be called in the batch process (it will
        synchronously invoke an indexer method which may block or cause a
        segfault).
        """


    def remove(document):
        """
        Remove the given document from this index.

        This method may be called from any process.
        """



class IFulltextIndexable(Interface):
    """
    Something which can be indexed for later search.
    """
    def uniqueIdentifier():
        """
        @return: a C{str} uniquely identifying this item.
        """


    def textParts():
        """
        @return: an iterable of unicode strings to be indexed as the text of
        this item.
        """


    def keywordParts():
        """
        @return: a C{dict} mapping C{str} to C{unicode} of additional
        metadata.  It will be possible to search on these fields using
        L{ISearchAggregator.search}.
        """


    def documentType():
        """
        @return: a C{str} uniquely identifying the type of this item.  Like
        the return value of L{keywordParts}, it will be possible to search
        for this using the C{"documentType"} key in the C{keywords} argument
        to L{ISearchAggregator.search}.
        """


    def sortKey():
        """
        @return: A unicode string that will be used as the key when sorting
        search results comprised of items of this type.
        """




class ISiteURLGenerator(Interface):
    """
    Lowest-level APIs for generating URLs which refer to parts of a Mantissa
    site.
    """
    def cleartextRoot(hostname=None):
        """
        Return the HTTP URL which is at the root of this site.

        @param hostname: An optional unicode string which, if specified, will
            be used as the hostname in the resulting URL, regardless of other
            considerations.

        @rtype: L{nevow.url.URL}
        """


    def encryptedRoot(hostname=None):
        """
        Return the HTTPS URL which is at the root of this site.

        @param hostname: An optional unicode string which, if specified, will
            be used as the hostname in the resulting URL, regardless of other
            considerations.

        @rtype: L{nevow.url.URL}
        """


    def rootURL(request):
        """
        Return the URL for the root of this website which is appropriate to use
        in links generated in response to the given request.

        @type request: L{twisted.web.http.Request}
        @param request: The request which is being responded to.

        @rtype: L{URL}
        @return: The location at which the root of the resource hierarchy for
            this website is available.

        NOTE: This function may take an URL (which is the "base" URL, relative
        to which all links in the ultimate view context will be interpreted)
        instead of a Request in the future.
        """



class IStaticShellContent(Interface):
    """
    Represents per-store header/footer content that's used to buttress
    the shell template
    """

    def getHeader():
        """
        Returns stan to be added to the page header.  Can return None
        if no header is desired.
        """

    def getFooter():
        """
        Returns stan to be added to the page footer.  Can return None
        if no footer is desired.
        """



class IViewer(Interface):
    def roleIn(userStore):
        """
        Retrieve a L{xmantissa.sharing.Role} object for the user that this
        viewer represents in the provided user-store.

        @param userStore: a store that contains some sharing roles.

        @type userStore: L{axiom.store.Store}

        @rtype: L{xmantissa.sharing.Role}
        """



class IWebViewer(IViewer):
    """
    An object that provides navigation bits for web content produced by
    Mantissa applications.
    """

    def wrapModel(model):
        """
        Converts application-provided model objects to L{IResource} providers.

        @param model: An L{Item} or L{SharedProxy}.
        """



class ISiteRootPlugin(Interface):
    """
    Plugin Interface for functionality provided at the root of the website.

    This interface is queried for on the Store by website.WebSite when
    processing an HTTP request.  Things which are installed on a Store using
    s.powerUp(x, ISiteRootPlugin) will be visible to individual users when
    installed on a user's store or visible to the general public when installed
    on a top-level store.
    """

    def resourceFactory(segments):
        """
        This is deprecated; implement L{produceResource} instead.

        Get an object that provides IResource.

        @type segments: list of str, representing decoded requested URL
        segments

        @return: None or a two-tuple of the IResource provider and the segments
        to pass to its locateChild.
        """
    del resourceFactory         # It's deprecated, so when we verifyObject, we
                                # don't want to check for this.


    def produceResource(request, segments, webViewer):
        """
        Get an object that provides IResource.

        @param request: An L{IRequest}.

        @type segments: list of str, representing decoded requested URL
        segments

        @param webViewer: An L{IWebViewer}.

        @return: None or a two-tuple of the IResource provider and the segments
        to pass to its locateChild.
        """


class IMantissaSite(Interface):
    # XXX this documentation is terrible, rephrase to describe something
    # abstract.
    """
    This is different from ISiteRootPlugin because it is invoked in a different
    context.  ISiteRootPlugin is a plugin powerup interface that lots of
    different things can provide.  This is an interface that only the site
    needs to provide, for L{CustomizedPublicPage} to do stuff with.
    """

    def siteProduceResource(request, segments, webViewer):
        """
        Give me a resource based on a bunch of ISiteRootPlugin powerups.
        """



class ISessionlessSiteRootPlugin(Interface):
    """
    L{ISessionlessSiteRootPlugin} powerups installed on a site store are
    powerups which can produce a resource to respond to a particular request,
    even if the browser in question has no session.

    This powerup interface exists mainly to allow applications to provide
    resources which are not subject to the redirect that L{nevow.guard}
    introduces.  This can be important for interacting with limited user-agents
    that do not support cookies or redirects.

    However, this is a temporary workaround, as L{nevow.guard} should have this
    redirect requirement removed.  See ticket #2494 for details.
    """

    def sessionlessProduceResource(request, segments):
        """
        Return a 2-tuple of C{(resource, segments)} if a resource can be found
        to match this request and its segments, or None.
        """


    def resourceFactory(segments):
        """
        This is deprecated; implement L{sessionlessProduceResource} instead.

        Get an object that provides IResource.

        @type segments: list of str, representing decoded requested URL
        segments

        @return: None or a two-tuple of the IResource provider and the segments
        to pass to its locateChild.
        """
    del resourceFactory         # It's deprecated, so when we verifyObject, we
                                # don't want to check for this.


class ICustomizable(Interface):
    """
    Factory for creating IResource objects which can be customized for
    a specific user.
    """
    def customizeFor(avatarName):
        """
        Retrieve a IResource provider specialized for the given avatar.

        @type avatarName: C{unicode}
        @param avatarName: The user for whom to return a specialized resource.

        @rtype: C{IResource}
        @return: A public-page resource, possibly customized for the
        indicated user.
        """

class ICustomizablePublicPage(Interface):
    """
    Don't use this.  Delete it if you notice it still exists but
    upgradePublicWeb2To3 has been removed.
    """

class IWebTranslator(Interface):
    """
    Provide methods for naming objects on the web, and vice versa.
    """

    def fromWebID(webID):
        """
        @param webID: A string that identifies an item through this translator.

        @return: an Item, or None if no Item is found.
        """

    def toWebID(item):
        """
        @param item: an item in the same store as this translator.

        @return: a string, shorter than 80 characters, which is an opaque
        identifier that may be used to look items up through this translator
        using fromWebID (or the legacy 'linkFrom')
        """


    def linkTo(storeID):
        """
        @param storeID: The Store ID of an Axiom item.

        @rtype: C{str}
        @return: An URL which refers to the item with the given Store ID.
        """

    def linkFrom(webID):
        """
        The inverse of L{linkTo}
        """

class INavigableElement(Interface):
    """Tab interface used by the web navigation plugin system.

    Plugins for this interface are retrieved when generating the navigation
    user-interface.  Each result has C{getTabs} invoked, after which the
    results are merged and the result used to construct various top- and
    secondary-level \"tabs\" which can be used to visit different parts of
    the application.
    """

    def getTabs():
        """Retrieve data about this elements navigation.

        This returns a list of C{xmantissa.webnav.Tab}s.

        For example, a powerup which wanted to install navigation under the
        Divmod tab would return this list:::

        [Tab("Divmod", self.storeID, 1.0
             children=[
                    Tab("Summary", self.storeID, 1.0),
                    Tab("Inbox", self.inbox.storeID, 0.8)
                    ])]
        """

class INavigableFragment(Interface):
    """
    Register an adapter to this interface in order to provide web UI content
    within the context of the 'private' application with navigation, etc.

    You will still need to produce some UI by implementing INavigableElement
    and registering a powerup for that as well, which allows users to navigate
    to this object.

    The primary requirement of this interface is that providers of it also
    provide L{nevow.inevow.IRenderer}.  The easiest way to achieve this is to
    subclass L{nevow.page.Element}.
    """

    title = Attribute(
        """
        The title of this fragment, which will be used in the <title> tag of
        the page displaying it, or otherwise outside the content area of the
        fragment but associated with it.
        """)

    fragmentName = Attribute(
        """
        The name of this fragment; a string used to look up the template from
        the current theme(s).  This is done by implementors of
        L{IWebViewer.wrapModel}.

        This attribute may be set to None or left unset if you do not want this
        type of customization.  However, if you do not set it, you must set a
        docFactory yourself, and your docFactory will not be changed to
        accommodate the user's preferred theme.
        """)

    docFactory = Attribute(
        """
        Nevow-style docFactory object.  Must be set if fragmentName is not.
        """)


    def head():
        """
        Provide some additional content to be included in the <head>
        section of the page when this fragment is being rendered.

        May return None if nothing needs to be added there.

        This method is optional.  If not implemented, nothing will be added to
        the head.
        """


    def locateChild(ctx, segments):
        """
        INavigableFragments may optionally provide a locateChild method similar
        to the one found on L{nevow.inevow.IResource.locateChild}.  You may
        implement this method if your INavigableFragment contains any resources
        which it may need to refer to with hyperlinks when rendered.  Please
        note that an INavigableFragment may be rendered on any page within an
        application, and that hyperlinks to resources returned from this method
        must always be to /private/<your-webid>/..., not the current page's
        URL, if you are using the default
        L{xmantissa.webapp.PrivateApplication} URL dispatcher.

        (There is a slight bug in the calling code's handling of Deferreds.
        If you wish to delegate to normal child-resource handling, you must
        return rend.NotFound exactly, not a Deferred which fires it.)
        """


    def setFragmentParent(fragmentParent):
        """
        Sets the L{LiveFragment} (or L{LivePage}) which is the logical parent
        of this fragment.

        See L{nevow.athena._LiveMixin.setFragmentParent}'s docstring for more
        information.
        """


    def customizeFor(userID):
        """
        This method is optional.  If not provided, it is the same as returning
        'self'.

        When a logged-in user is viewing an L{INavigableFragment} provider,
        this method will be invoked with that user's ID, and the returned
        navigable fragment will be presented to the user instead.

        @param userID: a user@host formatted string, indicating what user is
        viewing this.

        @type userID: L{unicode}

        @return: a fragment customized for the provided user-ID.

        @rtype: L{INavigableFragment}
        """



class ITab(Interface):
    """
    Abstract, non-UI representation of a tab that shows up in the UI.  The only
    concrete representation is xmantissa.webnav.Tab
    """



class IMessageReceiver(Interface):
    """
    An L{IMessageReceiver} is an object that can receive messages via
    inter-store messaging, L{xmantissa.interstore}.  Share an item with this
    interface and it will be able to receive messages queued with
    L{xmantissa.interstore.MessageQueue.queueMessage}.
    """

    def messageReceived(value, sender, target):
        """
        @param value: A value to be sent as the body of the message.
        @type value: L{xmantissa.interstore.Value}

        @param sender: a L{xmantissa.sharing.Identifier}, identifying the sender
        of the message being received.

        @param target: a L{xmantissa.sharing.Identifier}, identifying the
        object receiving the message, i.e. self.

        @return: a value for the response.
        @rtype: L{xmantissa.interstore.Value}
        """



class IDeliveryConsequence(Interface):
    """
    A provider of L{IDeliveryConsequence} can receive notifications of messages
    being successfully handled by a remote store via
    L{IMessageReceiver.messageReceived}.

    Providers of this interface must also be L{Item}s in Axiom stores so that
    they can persist between process invocations along with the queued message
    it is waiting for a response to.
    """

    def answerReceived(value, originalValue, originalSender, originalTarget):
        """
        An answer was received to a message sent via
        L{xmantissa.messaging.MessageQueue.queueMessage} with this
        L{IDeliveryConsequence} provider as its consequence.

        @param value: the value of the answer.
        @type value: L{xmantissa.interstore.Value}

        @param originalValue: the C{value} argument passed to the
        original L{IMessageReceiver.messageReceived} that this is a response
        to.
        @type originalData: L{xmantissa.interstore.Value}

        @param originalSender: the C{sender} argument passed to the original
        L{IMessageReceiver.messageReceived} that this is a response to.

        @param originalTarget: the C{target} argument passed to the original
        L{IMessageReceiver.messageReceived} that this is a response to.
        """



class IMessageRouter(Interface):
    """
    An L{IMessageRouter} is an object that can route messages between different
    users.  No guarantees are provided; all message routing is potentially
    unreliable.

    The suggested API for applications is
    L{xmantissa.interstore.AMPMessenger.messageRemote}.  Applications with
    specialized serialization needs might use the medium-level
    L{xmantissa.messaging.MessageQueue.queueMessage} instead.

    It is unlikely that you will want to use this interface unless you are
    implementing your own routing mechanism.  Any user of this interface should
    take care to test the failure cases, since most transports which implement
    this interface will, in practice, be (statistically speaking) extremely
    reliable and fail only in the most obscure cases.

    Application code should always be using something higher level, since this
    interface provides no mechanism for transactionality guarantees, and
    different providers may only know how to route to a subset of all possible
    destinations.  For example, the implementation installed on a user store
    will only know how to route to that user.
    """

    def routeMessage(sender, target, value, messageID):
        """
        Route a message to the given target.

        @param sender: The description of the shared item that is the sender of
        the message.
        @type sender: L{xmantissa.sharing.Identifier}

        @param target: The description of the shared item that is the target
        of the message.
        @type target: L{xmantissa.sharing.Identifier}

        @param messageID: An identifier for the message, unique to a given
        sender.
        @type messageID: L{int}

        @param value: The value of the message to be delivered.
        @type value: L{xmantissa.interstore.Value}
        """


    def routeAnswer(originalSender, originalTarget, value, messageID):
        """
        Route an answer to a message previously queued to a particular user.

        @param originalSender: The original sender of the message; in this
        case, the target of the answer.
        @type originalSender: L{xmantissa.sharing.Identifier}

        @param originalTarget: The original target of the message; in this
        case; the target of the answer.
        @type originalTarget: L{xmantissa.sharing.Identifier}

        @param messageID: The unique identifier for the message, as passed to
        L{IMessageRouter.routeMessage}.
        @type messageID: L{int}

        @param value: The value of the answer to be delivered.  This is not the
        value of the original message, but a separate value describing the
        result of processing the message.
        @type value: L{xmantissa.interstore.Value}

        @return: a L{Deferred} which fires with None if the message is
        successfully delivered, or fails with L{MessageTransportError} if the
        message could not be delivered.
        """



class IBenefactor(Interface):
    """
    Make accounts for users and give them things to use.
    """

    def endow(ticket, avatar):
        """
        Make a user and return it.  Give the newly created user new powerups or
        other functionality.

        This is only called when the user has confirmed the email address
        passed in by receiving a message and clicking on the link in the
        provided email.
        """

    def deprive(ticket, avatar):
        """
        Remove the increment of functionality or privilege that we have previously
        bestowed upon the indicated avatar.
        """

class IBenefactorFactory(Interface):
    """A factory which describes and creates IBenefactor providers.
    """

    def dependencies():
        """
        Return an iterable of other IBenefactorFactory providers that this one
        depends upon, and must be installed before this one is invoked.
        """

    def parameters():
        """
        Return a description of keyword parameters to be passed to instantiate.

        @rtype: A list of 4-tuples.  The first element of each tuple
        is a keyword argument to L{instantiate}.  The second describes
        the type of prompt to present for this field.  The third is a
        one-argument callable will should be invoked with a string the
        user supplies and should return the value for this keyword
        argument.  The fourth is a description of the purpose of this
        keyword argument.
        """

    def instantiate(**kw):
        """
        Create an IBenefactor provider and return it.
        """


class IQ2QService(Interface):

    q2qPortNumber = Attribute(
        """
        The TCP port number on which to listen for Q2Q connections.
        """)

    inboundTCPPortNumber = Attribute(
        """
        The TCP port number on which to listen for Q2Q data connections.
        """)

    publicIP = Attribute(
        """
        Dotted-quad format string representing the IP address via
        which this service is exposed to the public internet.
        """)

    udpEnabled = Attribute(
        """
        A boolean indicating whether or not PTCP connections will be
        allowed or attempted.
        """)

    def listenQ2Q(fromAddress, protocolsToFactories, serverDescription):
        """
        @see: L{vertex.q2q.Q2QService.connectQ2Q}
        """

    def connectQ2Q(fromAddress, toAddress, protocolName, protocolFactory,
                   usePrivateCertificate=None, fakeFromDomain=None,
                   chooser=None):
        """
        @see: L{vertex.q2q.Q2QService.connectQ2Q}
        """

class IPreferenceCollection(Interface):
    """
    I am an item that groups preferences into logical chunks.
    """

    def getPreferences():
        """
        Returns a mapping of preference-name->preference-value.
        """

    def getSections():
        """
        Returns a sequence of INavigableFragments or None. These fragments
        will be displayed alongside preferences under this collection's
        settings group.
        """

    def getPreferenceAttributes():
        """
        Returns a sequence of L{xmantissa.liveform.Parameter} instances - one
        for each preference.  The names of the parameters should correspond
        to the attribute names of the preference attributes on this item.
        """

    def getTabs():
        """
        Like L{ixmantissa.INavigableElement.getTabs}, but for preference tabs
        """

class ITemporalEvent(Interface):
    """
    I am an event which happens at a particular time and has a specific duration.
    """

    startTime = Attribute("""
    An extime.Time.  The start-point of this event.
    """)

    endTime = Attribute("""
    An extime.Time.  The end-point of this event.
    """)


class IDateBook(Interface):
    """
    A source of L{IAppointment}s which have times associated with them.
    """

    def eventsBetween(startTime, endTime):
        """
        Retrieve events which overlap a particular range.

        @param startTime: an L{epsilon.extime.Time} that begins a range.
        @param endTime: an L{epsilon.extime.Time} that ends a range.

        @return: an iterable of L{ITemporalEvent} providers.
        """



class IContactType(Interface):
    """
    A means by which communication with a L{Person} might occur.  For example,
    a telephone number.
    """
    allowMultipleContactItems = Attribute("""
    C{bool} indicating whether more than one contact item of this type can be
    created of a particular L{Person}.
    """)

    def getParameters(contactInfoItem):
        """
        Return some liveform parameters, one for each piece of information that is
        needed to construct a contact info item of this type.

        If C{contactInfoItem} is supplied, implementations may return C{None}
        to indicate that the given contact item is not editable.

        @param contactInfoItem: An existing contact info item of this type, or
        C{None}.  If not C{None}, then the current values of the contact info
        type will be used to provide suitable defaults for the parameters that
        are returned.
        @type contactInfoItem: L{axiom.item.Item} subclass.

        @return: Some liveform parameters or C{None}.
        @rtype: C{NoneType} or C{list} of L{xmantissa.liveform.Parameter}.
        """


    def createContactItem(person, **parameters):
        """
        Create a new instance of this contact type for the given person.

        @type person: L{Person}
        @param person: The person to whom the contact item pertains.

        @param parameters: The form input key/value pairs as returned by the
            L{xmantissa.liveform.LiveForm} constructed from L{getParameters}'s
            parameter instances.

        @return: The created contact item or C{None} if one was not created for
            any reason.
        """


    def getContactItems(person):
        """
        Return an iterator of contact items created by this contact type for
        the given person.

        @type person: L{Person}
        @param person: The person to whom the contact item pertains.
        """


    def uniqueIdentifier():
        """
        Return a C{unicode} string which, for the lifetime of a single Python
        process, uniquely identifies this type of contact information.
        """


    def descriptiveIdentifier():
        """
        A descriptive name for this type of contact information.

        @rtype: C{unicode}
        """


    def getEditFormForPerson(person):
        """
        Return a L{LiveForm} which will allow the given person's contact items
        to be edited.

        @type person: L{xmantissa.people.Person}

        @rtype: L{xmantissa.liveform.LiveForm}
        """


    def editContactItem(contact, **parameters):
        """
        Update the given contact item to reflect the new parameters.

        @param **parameters: The form input key/value pairs, as produced by the
            L{LiveForm} returned by L{getEditFormForPerson}.
        """


    def getContactGroup(contactItem):
        """
        Return a L{xmantissa.people.ContactGroup} describing the group
        affinity of a contact item of the type created by this contact type,
        or C{None} if the item doesn't have an explicit group.

        @param contactItem: A contact item.

        @rtype: L{xmantissa.people.ContactGroup} or C{NoneType}
        """


    def getReadOnlyView(contact):
        """
        Return an L{IRenderer} which will display the given contact.
        """



class IPeopleFilter(Interface):
    """
    Object which collects L{Person} items into a logical group.
    """
    filterName = Attribute("""
    The name of this filter; something which describes the type of people
    included in its group.""")


    def getPeopleQueryComparison(store):
        """
        Return a query comparison describing the subset of people in the given
        store which are included in this group.

        @type store: L{axiom.store.Store}

        @rtype: L{axiom.iaxiom.IComparison}
        """



class IOrganizerPlugin(Interface):
    """
    Powerup which provides additional functionality to Mantissa People.
    Organizer plugins add support for new kinds of person data (for example,
    one Organizer plugin might add support for contact information: physical
    addresses, email addresses, telephone numbers, etc.  Another plugin might
    retrieve and aggregate blog posts, or provide an interface for configuring
    sharing permissions).
    """
    name = Attribute('The C{unicode} display name of this plugin.')

    def getContactTypes():
        """
        Return an iterator of L{IContactType} providers supplied by this
        plugin.
        """


    def getPeopleFilters():
        """
        Return an iterator of L{IPeopleFilter} providers supplied by this
        plugin.
        """


    def personCreated(person):
        """
        Called when a new L{Person} is created.
        """


    def personNameChanged(person, oldName):
        """
        Called after a L{Person} item's name has been changed.

        @type person: L{Person}
        @param person: The person whose name is being changed.

        @type oldName: C{unicode}
        @param oldName: The previous value of L{{Person.name}.
        """


    def contactItemCreated(contact):
        """
        Called when a new contact item is created.

        @param contact: The new contact item.  It may be any object returned by
            an L{IContactType.createContactItem} implementation.
        """


    def contactItemEdited(contact):
        """
        Called when an existing contact item has been edited.

        @param contact: The contact item.
        """


    def personalize(person):
        """
        Return some plugin-specific state for the given person.

        @param person: A L{xmantissa.person.Person} instance.

        @return: Something renderable by Nevow.
        """



class IPersonFragment(Interface):
    """
    Deprecated.  Nothing in Mantissa cares about this interface.
    """



class IOffering(Interface):
    """
    Describes a product, service, application, or other unit of functionality
    which can be added to a Mantissa server.
    """

    name = Attribute("""
    What it is called.
    """)

    description = Attribute("""
    What it is.
    """)

    siteRequirements = Attribute("""
    A list of 2-tuples of (interface, powerupClass) of Axiom Powerups which
    will be installed on the Site store when this offering is installed if the
    store cannot be adapted to the given interface.
    """)

    appPowerups = Attribute("""
    A list of Axiom Powerups which will be installed on the App store when this
    offering is installed.  May be None if no App store is required (in this
    case, none will be created).
    """)

    installablePowerups = Attribute("""
    A C{list} of three-tuples each of which gives the name, description, and
    powerup item class for a unit of functionality which this offering provides
    for installation on a user store.
    """)

    loginInterfaces = Attribute("""
    A list of 2-tuples of (interface, description) of interfaces
    implemented by avatars provided by this offering, and human
    readable descriptions of the service provided by logging into
    them. Used by the statistics reporting system to label graphs of
    login activity.
    """)

    themes = Attribute("""
    Sequence of L{xmantissa.webtheme.XHTMLDirectoryTheme} instances,
    constituting themes that belong to this offering
    """)

    staticContentPath = Attribute("""
    A L{FilePath<twisted.python.filepath.FilePath>} referring to the root of
    the static content hierarchy for this offering.  This directory will be
    served automatically by Mantissa at C{/static/<offering name>/}.

    May be C{None} if there is no static content.
    """)

    version = Attribute("""
    L{twisted.python.versions.Version} instance indicating the version of
    this offering.  If included, the Version's value will be displayed to
    users once the offering is installed.  Defaults to None.
    """)



class IOfferingTechnician(Interface):
    """
    Support installation, uninstallation, and inspection of offerings.
    """
    def getInstalledOfferingNames():
        """
        Return a C{list} of C{unicode} strings giving the names of all
        installed L{IOffering}s.
        """


    def getInstalledOfferings():
        """
        Return a mapping from the names of installed L{IOffering} plugins to
        the plugins themselves.
        """


    def installOffering(offering):
        """
        Install the given offering plugin using the given configuration.

        @type offering: L{IOffering}
        @param offering: The offering to install.

        @raise L{xmantissa.offering.OfferingAlreadyInstalled}: If an offering
            with the same name as C{offering} is already installed.

        @return: The C{InstalledOffering} item created.
        """



class ISignupMechanism(Interface):
    """
    Describe an Item which can be instantiated to add a means of
    signing up to a Mantissa server.
    """

    name = Attribute("""
    What it is called.
    """)

    description = Attribute("""
    What it does.
    """)

    itemClass = Attribute("""
    An Axiom Item subclass which will be instantiated and added to the
    site store when this signup mechanism is selected.  The class
    should implement L{ISessionlessSiteRootPlugin} or
    L{ISiteRootPlugin}.
    """)

    configuration = Attribute("""
    XXX EDOC ME
    """)



class IProtocolFactoryFactory(Interface):
    """
    Powerup interface for Items which can create Twisted protocol factories.
    """
    def getFactory():
        """
        Return a Twisted protocol factory.
        """


class IBoxReceiverFactory(Interface):
    """
    Powerup interface for Items which can create L{IBoxReceiver} providers to
    be made accessible via the standard Mantissa AMP server.
    """
    protocol = Attribute(
        """
        A short string describing the commands (ie, the protocol) provided by
        the L{IBoxReceiver} this implementation can create.

        It is B{strongly} recommended that this be a versioned, URI-style
        identifier, after the fashion of XML namespace specifiers.  For
        example, the first version of a Divmod, Inc.-provided chat protocol
        might use I{https://divmod.com/ns/funny-chat}.  As no format
        describing AMP command sets (or other protocols built on AMP) is yet
        defined, there is no requirement that this URI be resolvable to an
        actual resource.  Its purpose at this time is merely to be unique.
        """)

    def getBoxReceiver():
        """
        Return an L{IBoxReceiver} which will be hooked up to an AMP connection
        and have messages sent to it.
        """



class IParameter(Interface):
    """
    Description of a single variable which will take on a value from external
    input and be used to perform some calculation or action.

    For example, an HTML form is a collection of IParameters, most likely one
    per input tag.  When POSTed, each input supplies its text value as the
    external input to a corresponding IParameter provider and the resulting
    collection is used to respond to the POST somehow.

    NOTE: This interface is highly unstable and subject to grossly incompatible
    changes.
    """

    # XXX - These shouldn't be attributes of IParameter, I expect.  They are
    # both really view things.  Either they goes into the template which is
    # used for this parameter (as an explanation to a user what the parameter
    # is), or some code which creates the view supplies them as parameters to
    # that object (in which case, it's probably more of a unique identifier in
    # that view context for this parameter). -exarkun
    name = Attribute(
        """
        A short C{unicode} string uniquely identifying this parameter within
        the context of a collection of L{IParameter} providers.
        """)

    label = Attribute(
        """
        A short C{unicode} string uniquely identifying this parameter within
        the context of a collection of L{IParameter} providers.
        """)

    # XXX - Another thing which belongs on the view.  Who even says this will
    # be rendered to an HTML form?
    type = Attribute(
        """
        One of C{liveform.TEXT_INPUT}, C{liveform.PASSWORD_INPUT},
        C{liveform.TEXTAREA_INPUT}, C{liveform.FORM_INPUT},
        C{liveform.RADIO_INPUT}, or C{liveform.CHECKBOX_INPUT} indicating the
        kind of input interface which will be presented for this parameter.
        """)

    # XXX - This shouldn't be an attribute of IParameter.  It's intended to be
    # displayed to end users, it belongs in a template.
    description = Attribute(
        """
        A long C{unicode} string explaining the meaning or purpose of this
        parameter.  May be C{None} to provide the end user with an unpleasant
        experience.
        """)

    # XXX - At this level, a default should be a structured object, not a
    # unicode string.  There is presently no way to serialize a structured
    # object into the view, though, so we use unicode here.
    default = Attribute(
        """
        A C{unicode} string which will be initially presented in the view as
        the value for this parameter, or C{None} if no such value should be
        presented.
        """)


    def viewFactory(parameter, default):
        """
        @type view: L{IParameter} provider
        @param view: The parameter for which to create a view.

        @param default: An object to return if no view can be created for the
            given parameter.

        @rtype: L{IParameterView} provider
        """


    # XXX - This is most definitely a view thing.
    def compact():
        """
        Mutate the parameter so that when a view object is created for it, it
        is more visually compact than it would otherwise have been.
        """


    def fromInputs(inputs):
        """
        Extract the value for this parameter from the given submission
        dictionary and return a structured value for this parameter.
        """



class IParameterView(IRenderer):
    """
    View interface for an individual LiveForm parameter.
    """
    patternName = Attribute("""
    Short string giving the name of the pattern for this parameter view.  Must
    be one of C{'text'}, C{'password'}, C{'repeatable-form'} or C{'choice'}.
    """)

    def setDefaultTemplate(tag):
        """
        Called by L{xmantissa.liveform.LiveForm} to specify the default
        template for this view.

        @type tag: L{nevow.stan.Tag} or C{nevow.stan.Proto}
        """


class IPublicPage(Interface):
    """
    Only needed for schema compatibility. This interface should be deleted once
    Axiom gains the ability to remove interfaces from existing stores.
    """


class IOneTimePadGenerator(Interface):
    """
    An object which can generate single-use pads for authentication purposes.
    """
    def generateOneTimePad(userStore):
        """
        Generate a one-time pad for the user who lives in the given store.

        @param userStore: A user's store.
        @type userStore: L{axiom.store.Store}

        @rtype: C{str}
        """



class ITerminalServerFactory(Interface):
    """
    A factory for L{ITerminalProtocol} providers which can create objects to
    handle input from and produce output to a terminal interface.
    """
    name = Attribute(
        "A short, user-facing C{unicode} string which identifies the "
        "functionality provided by this factory.")

    def buildTerminalProtocol(shellViewer):
        """
        Create and return a new L{ITerminalProtocol} provider to handle
        interact with a user.

        @param shellViewer: An L{IViewer} provider
        """



__all__ = [
    'IColumn', 'ITemplateNameResolver', 'IPreferenceAggregator',
    'ISearchProvider', 'ISearchAggregator', 'IFulltextIndexer',
    'IFulltextIndexable', 'IStaticShellContent', 'ISiteRootPlugin',
    'ISessionlessSiteRootPlugin', 'ICustomizable',
    'ICustomizablePublicPage', 'IWebTranslator', 'INavigableElement',
    'INavigableFragment', 'ITab', 'IBenefactor', 'IBenefactorFactory',
    'IQ2QService', 'IPreferenceCollection', 'ITemporalEvent', 'IDateBook',
    'IOrganizerPlugin', 'IPersonFragment', 'IOffering', 'ISignupMechanism',
    'IProtocolFactoryFactory', 'IParameterView', 'IOneTimePadGenerator',
    'ITerminalServerFactory',
    ]




from axiom.test.historic.stubloader import saveStub
from axiom.item import Item
from axiom.attributes import integer
from axiom.batch import processor

from xmantissa.fulltext import HypeIndexer, XapianIndexer, PyLuceneIndexer


class StubItem(Item):
    """
    Place-holder.  Stands in as an indexable thing, but no instances of this
    will ever actually be created.
    """
    __module__ = 'xmantissa.test.historic.stub_remoteIndexer1to2'

    attribute = integer()


StubSource = processor(StubItem)


def createDatabase(s):
    """
    Create a batch processor for L{StubItem} instances and add it as a message
    source to an instance of each of the kinds of indexers we support.
    """
    source = StubSource(store=s)
    for cls in [HypeIndexer, XapianIndexer, PyLuceneIndexer]:
        indexer = cls(store=s)
        source.addReliableListener(indexer)



if __name__ == '__main__':
    saveStub(createDatabase, 7053)



# -*- test-case-name: xmantissa.test.historic.test_ticket1to2 -*-
"""
Stub for Ticket upgrade. Creates a Ticket using the normal signup
machinery.
"""
from axiom.test.historic.stubloader import saveStub
from xmantissa.signup import SignupConfiguration, FreeTicketSignup, Multifactor
from xmantissa.provisioning import BenefactorFactory
from xmantissa.webadmin import AdministrativeBenefactor

def createDatabase(s):
    mff = BenefactorFactory("", "", AdministrativeBenefactor)
    sc = SignupConfiguration(store=s)
    sc.installOn(s)
    s.parent = s
    signup = sc.createSignup(u'bob', FreeTicketSignup, {'prefixURL': u'/signup'},
                             {mff: {}}, None,
                             u'Sign Up')
    t = signup.booth.createTicket(signup, u'bob@example.com', signup.benefactor)

if __name__ == '__main__':
    saveStub(createDatabase, 10876)



import textwrap

from nevow import loaders, tags
from nevow.livetrial import testcase

from xmantissa import liveform

class TextInput(testcase.TestCase):
    jsClass = u'Mantissa.Test.Text'

    def getWidgetDocument(self):
        f = liveform.LiveForm(
            self.submit,
            [liveform.Parameter('argument',
                                liveform.TEXT_INPUT,
                                unicode,
                                'A text input field: ',
                                default=u'hello world')])
        f.setFragmentParent(self)
        return f


    def submit(self, argument):
        self.assertEquals(argument, u'hello world')



class MultiTextInput(testcase.TestCase):
    jsClass = u'Mantissa.Test.MultiText'

    def submit(self, sequence):
        self.assertEquals(sequence, [1, 2, 3, 4])


    def getWidgetDocument(self):
        f = liveform.LiveForm(
                self.submit,
                (liveform.ListParameter('sequence',
                                        int,
                                        4,
                                        'A bunch of text inputs: ',
                                        defaults=(1, 2, 3, 4)),))
        f.setFragmentParent(self)
        return f



class TextArea(testcase.TestCase):
    jsClass = u'Mantissa.Test.TextArea'

    defaultText = textwrap.dedent(u"""
    Come hither, sir.
    Though it be honest, it is never good
    To bring bad news. Give to a gracious message
    An host of tongues; but let ill tidings tell
    Themselves when they be felt.
    """).strip()

    def submit(self, argument):
        self.assertEquals(argument, self.defaultText)


    def getWidgetDocument(self):
        f = liveform.LiveForm(
            self.submit,
            [liveform.Parameter('argument',
                                liveform.TEXTAREA_INPUT,
                                unicode,
                                'A text area: ',
                                default=self.defaultText)])
        f.setFragmentParent(self)
        return f



class Select(testcase.TestCase):
    jsClass = u'Mantissa.Test.Select'

    def submit(self, argument):
        self.assertEquals(argument, u"apples")


    def getWidgetDocument(self):
        # XXX No support for rendering these yet!
        f = liveform.LiveForm(
            self.submit,
            [liveform.Parameter('argument', None, unicode)])
        f.docFactory = loaders.stan(tags.form(render=tags.directive('liveElement'))[
            tags.select(name="argument")[
                tags.option(value="apples")["apples"],
                tags.option(value="oranges")["oranges"]],
            tags.input(type='submit', render=tags.directive('submitbutton'))])
        f.setFragmentParent(self)
        return f



class Choice(testcase.TestCase):
    jsClass = u'Mantissa.Test.Choice'

    def submit(self, argument):
        self.assertEquals(argument, 2)


    def getWidgetDocument(self):
        f = liveform.LiveForm(
            self.submit,
            [liveform.ChoiceParameter('argument',
                [('One', 1, False),
                 ('Two', 2, True),
                 ('Three', 3, False)])])
        f.setFragmentParent(self)
        return f



class ChoiceMultiple(testcase.TestCase):
    jsClass = u'Mantissa.Test.ChoiceMultiple'

    def submit(self, argument):
        self.assertIn(1, argument)
        self.assertIn(3, argument)


    def getWidgetDocument(self):
        f = liveform.LiveForm(
            self.submit,
            [liveform.ChoiceParameter('argument',
                [('One', 1, True),
                 ('Two', 2, False),
                 ('Three', 3, True)],
                "Choosing multiples from a list.",
                multiple=True)])
        f.setFragmentParent(self)
        return f



SPECIAL = object() # guaranteed to fuck up JSON if it ever gets there by
                   # accident.

class Traverse(testcase.TestCase):
    jsClass = u'Mantissa.Test.Traverse'

    def submit(self, argument, group):
        self.assertEquals(argument, u'hello world')
        self.assertEquals(group, SPECIAL)


    def paramfilter(self, param1):
        self.assertEquals(param1, u'goodbye world')
        return SPECIAL


    def getWidgetDocument(self):
        f = liveform.LiveForm(
            self.submit,
            [liveform.Parameter('argument',
                                liveform.TEXT_INPUT,
                                unicode,
                                'A text input field: ',
                                default=u'hello world'),
             liveform.Parameter('group',
                                liveform.FORM_INPUT,
                                liveform.LiveForm(self.paramfilter,
                                                  [liveform.Parameter
                                                   ('param1',
                                                    liveform.TEXT_INPUT,
                                                    unicode,
                                                    'Another input field: ',
                                                    default=u'goodbye world')]),
                                'A form input group: ',
                                )])
        f.setFragmentParent(self)
        return f

class SetInputValues(testcase.TestCase):
    jsClass = u'Mantissa.Test.SetInputValues'

    def submit(self, choice, choiceMult, text, passwd, textArea, checkbox):
        """
        Assert that all input values have been reversed/inverted
        """
        self.assertEqual(choice, 1)
        self.assertEqual(choiceMult, (2, 3))
        self.assertEqual(text, 'dlrow olleh')
        self.assertEqual(passwd, 'yek terces')
        self.assertEqual(textArea, '2 dlrow olleh')
        self.failIf(checkbox)

    def getWidgetDocument(self):
        """
        Make a LiveForm with one of each kind of input, except for radio
        buttons, since with the current liveform support for them it's
        difficult to use them with a single form, and it's not so important to
        do anything else right now
        """
        f = liveform.LiveForm(
                self.submit,
                (liveform.ChoiceParameter(
                    'choice',
                    (('0', 0, True), ('1', 1, False))),
                 liveform.ChoiceParameter(
                     'choiceMult',
                     (('0', 0, True),  ('1', 1, True),
                      ('2', 2, False), ('3', 3, False)),
                     multiple=True),
                 liveform.Parameter(
                    'text',
                    liveform.TEXT_INPUT,
                    unicode,
                    default=u'hello world'),
                 liveform.Parameter(
                    'passwd',
                    liveform.PASSWORD_INPUT,
                    unicode,
                    default=u'secret key'),
                 liveform.Parameter(
                     'textArea',
                     liveform.TEXTAREA_INPUT,
                     unicode,
                     default=u'hello world 2'),
                 liveform.Parameter(
                     'checkbox',
                     liveform.CHECKBOX_INPUT,
                     bool,
                     default=True)))
        f.setFragmentParent(self)
        return f

class FormName(testcase.TestCase):
    """
    Test that the form name is correctly set client-side
    """
    jsClass = u'Mantissa.Test.FormName'

    def getWidgetDocument(self):
        """
        Make a nested form
        """
        f = liveform.LiveForm(
                lambda **k: None,
                (liveform.Parameter(
                    'inner-form',
                    liveform.FORM_INPUT,
                    liveform.LiveForm(
                        lambda **k: None,
                        (liveform.Parameter(
                            'inner-parameter',
                            liveform.TEXT_INPUT,
                            unicode,
                            ''),), ())),))
        f.setFragmentParent(self)
        return f




"""
Create an L{AnonymousSite} in a database by itself.
"""

from axiom.test.historic.stubloader import saveStub
from axiom.dependency import installOn

from xmantissa.publicweb import AnonymousSite

def createDatabase(s):
    installOn(AnonymousSite(store=s), s)

if __name__ == '__main__':
    saveStub(createDatabase, 16560)



# -*- test-case-name: xmantissa.test.test_liveform -*-

"""

XXX HYPER TURBO SUPER UNSTABLE DO NOT USE XXX

"""

import warnings

from zope.interface import implements

from twisted.python.components import registerAdapter
from twisted.internet.defer import maybeDeferred, gatherResults

from epsilon.structlike import record

from nevow import inevow, tags, page, athena
from nevow.athena import expose
from nevow.page import Element, renderer
from nevow.loaders import stan

from xmantissa import webtheme
from xmantissa.fragmentutils import PatternDictionary, dictFillSlots
from xmantissa.ixmantissa import IParameter, IParameterView


_LIVEFORM_JS_CLASS = u'Mantissa.LiveForm.FormWidget'
_SUBFORM_JS_CLASS = u'Mantissa.LiveForm.SubFormWidget'



class InputError(athena.LivePageError):
    """
    Base class for all errors related to rejected input values.
    """
    jsClass = u'Mantissa.LiveForm.InputError'



TEXT_INPUT = 'text'
PASSWORD_INPUT = 'password'
TEXTAREA_INPUT = 'textarea'
FORM_INPUT = 'form'
RADIO_INPUT = 'radio'
CHECKBOX_INPUT = 'checkbox'



class _SelectiveCoercer(object):
    """
    Mixin defining a L{IParameter.fromInputs} implementation which extracts the
    input associated with this parameter based on C{self.name} and passes it to
    C{self.coercer}, returning the result.
    """
    def fromInputs(self, inputs):
        """
        Extract the inputs associated with the child forms of this parameter
        from the given dictionary and coerce them using C{self.coercer}.

        @type inputs: C{dict} mapping C{unicode} to C{list} of C{unicode}
        @param inputs: The contents of a form post, in the conventional
            structure.

        @rtype: L{Deferred}
        @return: The structured data associated with this parameter represented
            by the post data.
        """
        try:
            values = inputs[self.name]
        except KeyError:
            raise ConfigurationError(
                "Missing value for input: " + self.name)
        return self.coerceMany(values)


    def coerceMany(self, values):
        """
        Convert the given C{list} of C{unicode} inputs to structured data.

        @param values: The inputs associated with this parameter's name in the
            overall inputs mapping.
        """
        return self.coercer(values[0])



class Parameter(record('name type coercer label description default '
                       'viewFactory',
                       label=None,
                       description=None,
                       default=None,
                       viewFactory=IParameterView), _SelectiveCoercer):
    """
    @type name: C{unicode}
    @ivar name: A name uniquely identifying this parameter within a particular
        form.

    @ivar type: One of C{TEXT_INPUT}, C{PASSWORD_INPUT}, C{TEXTAREA_INPUT},
        C{RADIO_INPUT}, or C{CHECKBOX_INPUT} indicating the kind of input
        interface which will be presented for this parameter.

    @type description: C{unicode} or C{NoneType}
    @ivar description: An explanation of the meaning or purpose of this
        parameter which will be presented in the view, or C{None} if the user
        is intended to guess.

    @type default: C{unicode} or C{NoneType}
    @ivar default: A value which will be initially presented in the view as the
        value for this parameter, or C{None} if no such value is to be
        presented.

    @ivar viewFactory: A two-argument callable which returns an
        L{IParameterView} provider which will be used as the view for this
        parameter, if one can be provided.  It will be invoked with the
        parameter as the first argument and a default value as the second
        argument.  The default should be returned if no view can be provided
        for the given parameter.
    """
    implements(IParameter)

    def __init__(self, *a, **kw):
        super(Parameter, self).__init__(*a, **kw)
        if self.type == FORM_INPUT:
            warnings.warn(
                "Create a FormParameter, not a Parameter with type FORM_INPUT",
                category=DeprecationWarning,
                stacklevel=2)


    def compact(self):
        """
        Compact FORM_INPUTs by calling their C{compact} method.  Don't do
        anything for other types of input.
        """
        if self.type == FORM_INPUT:
            self.coercer.compact()


    def clone(self, default):
        """
        Make a copy of this parameter, supplying a different default.

        @type default: C{unicode} or C{NoneType}
        @param default: A value which will be initially presented in the view
        as the value for this parameter, or C{None} if no such value is to be
        presented.

        @rtype: L{Parameter}
        """
        return self.__class__(
            self.name,
            self.type,
            self.coercer,
            self.label,
            self.description,
            default,
            self.viewFactory)



class FormParameter(record('name form label description default viewFactory',
                           label=None, description=None, default=None,
                           viewFactory=IParameterView),
                    _SelectiveCoercer):
    """
    A parameter which is a collection of other parameters, as composed by a
    L{LiveForm}.

    @type name: C{unicode}
    @ivar name: A name uniquely identifying this parameter within a
    particular form.

    @type form: L{LiveForm}
    @ivar form: The form which defines the grouped parameters.

    @type description: C{unicode} or C{NoneType}
    @ivar description: An explanation of the meaning or purpose of this
        parameter which will be presented in the view, or C{None} if the user
        is intended to guess.

    @type default: C{unicode} or C{NoneType}
    @ivar default: A value which will be initially presented in the view as the
        value for this parameter, or C{None} if no such value is to be
        presented.

    @ivar viewFactory: A two-argument callable which returns an
        L{IParameterView} provider which will be used as the view for this
        parameter, if one can be provided.  It will be invoked with the
        parameter as the first argument and a default value as the second
        argument.  The default should be returned if no view can be provided
        for the given parameter.
    """
    implements(IParameter)

    type = FORM_INPUT

    def compact(self):
        """
        Compact the wrapped form.
        """
        self.form.compact()


    def coercer(self, value):
        """
        Invoke the wrapped form with the given value and return its result.
        """
        return self.form.invoke(value)



class CreateObject(record('values setter')):
    """
    Represent one object which should be created as a result of a submission to
    a L{ListChangeParameter}.

    @ivar values: The coerced data from the submission which should be used to
        create this object.

    @ivar setter: A one-argument callable which must be called with the created
        object once it is created.  Until this is called, the
        L{ListChangeParameter} will be in a state where it
        will not handle submissions correctly. (XXX - This could be hooked up
        to a Deferred, and L{ListChangeParameter}'s C{coercer}
        method could gatherResults() on all of these, delaying the success of
        the submission until all created objects have been set).
    """



class EditObject(record('object values')):
    """
    Represent changes to be made to an object as the result of the submission
    of a L{ListChangeParameter}.

    @ivar object: The object which is to be edited.  This is one of the
        elements of the C{modelObjects} sequence passed to
        L{ListChangeParameter.__init__} or it is one of the
        objects subsequently added to the L{ListChangeParameter} by
        a call to L{CreateObject.setter}.

    @ivar values: The new values for this object from the submission.
    """
    def __cmp__(self, other):
        return cmp((self.object, self.values), (other.object, other.values))



class ListChanges(record('create edit delete')):
    """
    Represent the submission of a L{ListChangeParameter}.

    @ivar create: A list of L{CreateObject} instances, one for each new object
        to be created as a result of the submission.

    @ivar edit: A list of L{EditObject} instances, one for each existing object
        modified by the submission.

    @ivar delete: A list of model objects which should be deleted as a result
        of the submission.
    """



class ListChangeParameter(record('name parameters defaults modelObjects '
                                 'modelObjectDescription viewFactory',
                                 defaults=(), modelObjects=(),
                                 modelObjectDescription=u'',
                                 viewFactory=IParameterView),
                          _SelectiveCoercer):
    """
    Use this parameter if you want to render lists of objects as forms, and to
    allow the objects to be edited and deleted, as well as to have new objects
    created.

    Parameters of this type will be coerced in L{ListChanges} objects.

    @type name: C{unicode}
    @ivar name: A name uniquely identifying this parameter within a
    particular form.

    @type parameters: C{list}
    @ivar parameters: sequence of L{Parameter} instances, describing the
    contents of the repeatable form.

    @type defaults: C{list}
    @ivar defaults: A sequence of dictionaries mapping names of L{parameters}
        to values.

    @type modelObjects: C{list}
    @ivar modelObjects: A sequence of opaque objects, one for each item in
    C{defaults}.

    @type modelObjectDescription: C{unicode}
    @param modelObjectDescription: A description of the type of model object
    being edited, e.g. 'Email Address'.  Defaults to the empty string.

    @type viewFactory: callable
    @ivar viewFactory: A two-argument callable which returns an
        L{IParameterView} provider which will be used as the view for this
        parameter, if one can be provided.  It will be invoked with the
        parameter as the first argument and a default value as the second
        argument.  The default should be returned if no view can be provided
        for the given parameter.
    """
    _parameterIsCompact = False
    type = None

    _IDENTIFIER_KEY = u'__repeated-liveform-id__'
    _NO_OBJECT_MARKER = object()

    def __init__(self, *a, **k):
        super(ListChangeParameter, self).__init__(*a, **k)
        self.liveFormFactory = LiveForm
        self.repeatedLiveFormWrapper = RepeatedLiveFormWrapper
        self._idsToObjects = {}
        self._lastValues = {}
        self._defaultStuff = []
        for (defaultObject, defaultValues) in zip(self.modelObjects, self.defaults):
            identifier = self._idForObject(defaultObject)
            self._lastValues[identifier] = defaultValues
            self._defaultStuff.append((defaultValues, identifier))


    def compact(self):
        """
        Remember whether we were compacted or not, so we can relay this to our
        view.
        """
        self._parameterIsCompact = True


    def _prepareSubForm(self, liveForm):
        """
        Utility for turning liveforms into subforms, and compacting them as
        necessary.

        @param liveForm: a liveform.
        @type liveForm: L{LiveForm}

        @return: a sub form.
        @rtype: L{LiveForm}
        """
        liveForm = liveForm.asSubForm(self.name) # XXX Why did this work???
        # if we are compact, tell the liveform so it can tell its parameters
        # also
        if self._parameterIsCompact:
            liveForm.compact()
        return liveForm


    def _cloneDefaultedParameter(self, original, default):
        """
        Make a copy of the parameter C{original}, supplying C{default} as the
        default value.

        @type original: L{Parameter} or L{ChoiceParameter}
        @param original: A liveform parameter.

        @param default: An alternate default value for the parameter.

        @rtype: L{Parameter} or L{ChoiceParameter}
        @return: A new parameter.
        """
        if isinstance(original, ChoiceParameter):
            default = [Option(o.description, o.value, o.value in default)
                        for o in original.choices]
        return original.clone(default)


    _counter = 0
    def _allocateID(self):
        """
        Allocate an internal identifier.

        @rtype: C{int}
        """
        self._counter += 1
        return self._counter


    def _idForObject(self, defaultObject):
        """
        Generate an opaque identifier which can be used to talk about
        C{defaultObject}.

        @rtype: C{int}
        """
        identifier = self._allocateID()
        self._idsToObjects[identifier] = defaultObject
        return identifier


    def _objectFromID(self, identifier):
        """
        Find the object associated with the identifier C{identifier}.

        @type identifier: C{int}
        """
        return self._idsToObjects[identifier]


    def _newIdentifier(self):
        """
        Make a new identifier for an as-yet uncreated model object.

        @rtype: C{int}
        """
        id = self._allocateID()
        self._idsToObjects[id] = self._NO_OBJECT_MARKER
        self._lastValues[id] = None
        return id


    def _makeALiveForm(self, parameters, identifier, removable=True):
        """
        Make a live form with the parameters C{parameters}, which will be used
        to edit the values/model object with identifier C{identifier}.

        @type parameters: C{list}
        @param parameters: list of L{Parameter} instances.

        @type identifier: C{int}

        @type removable: C{bool}

        @rtype: L{repeatedLiveFormWrapper}
        """
        liveForm = self.liveFormFactory(lambda **k: None, parameters, self.name)
        liveForm = self._prepareSubForm(liveForm)
        liveForm = self.repeatedLiveFormWrapper(liveForm, identifier, removable)
        liveForm.docFactory = webtheme.getLoader(liveForm.fragmentName)
        return liveForm


    def _makeDefaultLiveForm(self, (defaults, identifier)):
        """
        Make a liveform suitable for editing the set of default values C{defaults}.

        @type defaults: C{dict}
        @param defaults: Mapping of parameter names to values.

        @rtype: L{repeatedLiveFormWrapper}
        """
        parameters = [self._cloneDefaultedParameter(p, defaults[p.name])
                        for p in self.parameters]
        return self._makeALiveForm(parameters, identifier)


    def getInitialLiveForms(self):
        """
        Make and return as many L{LiveForm} instances as are necessary to hold
        our default values.

        @return: some subforms.
        @rtype: C{list} of L{LiveForm}
        """
        liveForms = []
        if self._defaultStuff:
            for values in self._defaultStuff:
                liveForms.append(self._makeDefaultLiveForm(values))
        else:
            # or only one, for the first new thing
            liveForms.append(
                self._makeALiveForm(
                    self.parameters, self._newIdentifier(), False))
        return liveForms


    def asLiveForm(self):
        """
        Make and return a form, using L{parameters}.

        @return: a sub form.
        @rtype: L{LiveForm}
        """
        return self._makeALiveForm(self.parameters, self._newIdentifier())


    def _coerceSingleRepetition(self, dataSet):
        """
        Make a new liveform with our parameters, and get it to coerce our data
        for us.
        """
        # make a liveform because there is some logic in _coerced
        form = LiveForm(lambda **k: None, self.parameters, self.name)
        return form.fromInputs(dataSet)


    def _extractCreations(self, dataSets):
        """
        Find the elements of C{dataSets} which represent the creation of new
        objects.

        @param dataSets: C{list} of C{dict} mapping C{unicode} form submission
            keys to form submission values.

        @return: iterator of C{tuple}s with the first element giving the opaque
            identifier of an object which is to be created and the second
            element giving a C{dict} of all the other creation arguments.
        """
        for dataSet in dataSets:
            modelObject = self._objectFromID(dataSet[self._IDENTIFIER_KEY])
            if modelObject is self._NO_OBJECT_MARKER:
                dataCopy = dataSet.copy()
                identifier = dataCopy.pop(self._IDENTIFIER_KEY)
                yield identifier, dataCopy


    def _extractEdits(self, dataSets):
        """
        Find the elements of C{dataSets} which represent changes to existing
        objects.

        @param dataSets: C{list} of C{dict} mapping C{unicode} form submission
            keys to form submission values.

        @return: iterator of C{tuple}s with the first element giving a model
            object which is being edited and the second element giving a
            C{dict} of all the other arguments.
        """
        for dataSet in dataSets:
            modelObject = self._objectFromID(dataSet[self._IDENTIFIER_KEY])
            if modelObject is not self._NO_OBJECT_MARKER:
                dataCopy = dataSet.copy()
                identifier = dataCopy.pop(self._IDENTIFIER_KEY)
                yield identifier, dataCopy


    def _coerceAll(self, inputs):
        """
        XXX
        """
        def associate(result, obj):
            return (obj, result)

        coerceDeferreds = []
        for obj, dataSet in inputs:
            oneCoerce = self._coerceSingleRepetition(dataSet)
            oneCoerce.addCallback(associate, obj)
            coerceDeferreds.append(oneCoerce)
        return gatherResults(coerceDeferreds)


    def coercer(self, dataSets):
        """
        Coerce all of the repetitions and sort them into creations, edits and
        deletions.

        @rtype: L{ListChanges}
        @return: An object describing all of the creations, modifications, and
            deletions represented by C{dataSets}.
        """
        # Xxx - This does a slightly complex (hey, it's like 20 lines, how
        # complex could it really be?) thing to figure out which elements are
        # newly created, which elements were edited, and which elements no
        # longer exist.  It might be simpler if the client kept track of this
        # and passed a three-tuple of lists (or whatever - some separate data
        # structures) to the server, so everything would be all figured out
        # already.  This would require the client
        # (Mantissa.LiveForm.RepeatableForm) to be more aware of what events
        # the user is triggering in the browser so that it could keep state for
        # adds/deletes/edits separately from DOM and widget objects.  This
        # would remove the need for RepeatedLiveFormWrapper.
        def makeSetter(identifier, values):
            def setter(defaultObject):
                self._idsToObjects[identifier] = defaultObject
                self._lastValues[identifier] = values
            return setter

        created = self._coerceAll(self._extractCreations(dataSets))
        edited = self._coerceAll(self._extractEdits(dataSets))

        coerceDeferred = gatherResults([created, edited])
        def cbCoerced((created, edited)):
            receivedIdentifiers = set()

            createObjects = []
            for (identifier, dataSet) in created:
                receivedIdentifiers.add(identifier)
                createObjects.append(
                    CreateObject(dataSet, makeSetter(identifier, dataSet)))

            editObjects = []
            for (identifier, dataSet) in edited:
                receivedIdentifiers.add(identifier)
                lastValues = self._lastValues[identifier]
                if dataSet != lastValues:
                    modelObject = self._objectFromID(identifier)
                    editObjects.append(EditObject(modelObject, dataSet))
                    self._lastValues[identifier] = dataSet

            deleted = []
            for identifier in set(self._idsToObjects) - receivedIdentifiers:
                existing = self._objectFromID(identifier)
                if existing is not self._NO_OBJECT_MARKER:
                    deleted.append(existing)
                self._idsToObjects.pop(identifier)

            return ListChanges(createObjects, editObjects, deleted)

        coerceDeferred.addCallback(cbCoerced)
        return coerceDeferred



MULTI_TEXT_INPUT = 'multi-text'

class ListParameter(record('name coercer count label description defaults '
                           'viewFactory',
                           label=None,
                           description=None,
                           defaults=None,
                           viewFactory=IParameterView)):

    type = MULTI_TEXT_INPUT
    def compact(self):
        """
        Don't do anything.
        """


    def fromInputs(self, inputs):
        """
        Extract the inputs associated with this parameter from the given
        dictionary and coerce them using C{self.coercer}.

        @type inputs: C{dict} mapping C{str} to C{list} of C{str}
        @param inputs: The contents of a form post, in the conventional
            structure.

        @rtype: L{Deferred}
        @return: A Deferred which will be called back with a list of the
            structured data associated with this parameter.
        """
        outputs = []
        for i in xrange(self.count):
            name = self.name + '_' + str(i)
            try:
                value = inputs[name][0]
            except KeyError:
                raise ConfigurationError(
                    "Missing value for field %d of %s" % (i, self.name))
            else:
                outputs.append(maybeDeferred(self.coercer, value))
        return gatherResults(outputs)



CHOICE_INPUT = 'choice'
MULTI_CHOICE_INPUT = 'multi-choice'


class Option(record('description value selected')):
    """
    A single choice for a L{ChoiceParameter}.
    """


class ChoiceParameter(record('name choices label description multiple '
                             'viewFactory',
                             label=None,
                             description="",
                             multiple=False,
                             viewFactory=IParameterView), _SelectiveCoercer):
    """
    A choice parameter, represented by a <select> element in HTML.

    @ivar choices: A sequence of L{Option} instances (deprecated: a sequence of
        three-tuples giving the attributes of L{Option} instances).

    @ivar multiple: C{True} if multiple choice selections are allowed

    @ivar viewFactory: A two-argument callable which returns an
        L{IParameterView} provider which will be used as the view for this
        parameter, if one can be provided.  It will be invoked with the
        parameter as the first argument and a default value as the second
        argument.  The default should be returned if no view can be provided
        for the given parameter.
    """
    def __init__(self, *a, **kw):
        ChoiceParameter.__bases__[0].__init__(self, *a, **kw)
        if self.choices and isinstance(self.choices[0], tuple):
            warnings.warn(
                "Pass a list of Option instances to ChoiceParameter, "
                "not a list of tuples.",
                category=DeprecationWarning,
                stacklevel=2)
            self.choices = [Option(*o) for o in self.choices]


    def type(self):
        if self.multiple:
            return MULTI_CHOICE_INPUT
        return CHOICE_INPUT
    type = property(type)


    def coercer(self, value):
        if self.multiple:
            return tuple(self.choices[int(v)].value for v in value)
        return self.choices[int(value)].value


    def compact(self):
        """
        Don't do anything.
        """


    def clone(self, choices):
        """
        Make a copy of this parameter, supply different choices.

        @param choices: A sequence of L{Option} instances.
        @type choices: C{list}

        @rtype: L{ChoiceParameter}
        """
        return self.__class__(
            self.name,
            choices,
            self.label,
            self.description,
            self.multiple,
            self.viewFactory)



class ConfigurationError(Exception):
    """
    User-specified configuration for a newly created Item was invalid or
    incomplete.
    """



class InvalidInput(Exception):
    """
    Data entered did not meet the requirements of the coercer.
    """



def _legacySpecialCases(form, patterns, parameter):
    """
    Create a view object for the given parameter.

    This function implements the remaining view construction logic which has
    not yet been converted to the C{viewFactory}-style expressed in
    L{_LiveFormMixin.form}.

    @type form: L{_LiveFormMixin}
    @param form: The form fragment which contains the given parameter.
    @type patterns: L{PatternDictionary}
    @type parameter: L{Parameter}, L{ChoiceParameter}, or L{ListParameter}.
    """
    p = patterns[parameter.type + '-input-container']

    if parameter.type == TEXTAREA_INPUT:
        p = dictFillSlots(p, dict(label=parameter.label,
                                  name=parameter.name,
                                  value=parameter.default or ''))
    elif parameter.type == MULTI_TEXT_INPUT:
        subInputs = list()

        for i in xrange(parameter.count):
            subInputs.append(dictFillSlots(patterns['input'],
                                dict(name=parameter.name + '_' + str(i),
                                     type='text',
                                     value=parameter.defaults[i])))

        p = dictFillSlots(p, dict(label=parameter.label or parameter.name,
                                  inputs=subInputs))

    else:
        if parameter.default is not None:
            value = parameter.default
        else:
            value = ''

        if parameter.type == CHECKBOX_INPUT and parameter.default:
            inputPattern = 'checked-checkbox-input'
        else:
            inputPattern = 'input'

        p = dictFillSlots(
            p, dict(label=parameter.label or parameter.name,
                    input=dictFillSlots(patterns[inputPattern],
                                        dict(name=parameter.name,
                                             type=parameter.type,
                                             value=value))))

    p(**{'class' : 'liveform_'+parameter.name})

    if parameter.description:
        description = patterns['description'].fillSlots(
                           'description', parameter.description)
    else:
        description = ''

    return dictFillSlots(
        patterns['parameter-input'],
        dict(input=p, description=description))



class _LiveFormMixin(record('callable parameters description',
                            description=None)):
    jsClass = _LIVEFORM_JS_CLASS

    subFormName = None

    fragmentName = 'liveform'
    compactFragmentName = 'liveform-compact'

    def __init__(self, *a, **k):
        super(_LiveFormMixin, self).__init__(*a, **k)
        if self.docFactory is None:
            # Give subclasses a chance to assign their own docFactory.
            self.docFactory = webtheme.getLoader(self.fragmentName)


    def compact(self):
        """
        Switch to the compact variant of the live form template.

        By default, this will simply create a loader for the
        C{self.compactFragmentName} template and compact all of this form's
        parameters.
        """
        self.docFactory = webtheme.getLoader(self.compactFragmentName)
        for param in self.parameters:
            param.compact()


    def getInitialArguments(self):
        if self.subFormName:
            subFormName = self.subFormName.decode('utf-8')
        else:
            subFormName = None
        return (subFormName,)


    def asSubForm(self, name):
        """
        Make a form suitable for nesting within another form (a subform) out
        of this top-level liveform.

        @param name: the name of the subform within its parent.
        @type name: C{unicode}

        @return: a subform.
        @rtype: L{LiveForm}
        """
        self.subFormName = name
        self.jsClass = _SUBFORM_JS_CLASS
        return self


    def _getDescription(self):
        descr = self.description
        if descr is None:
            descr = self.callable.__name__
        return descr


    def submitbutton(self, request, tag):
        """
        Render an INPUT element of type SUBMIT which will post this form to the
        server.
        """
        return tags.input(type='submit',
                          name='__submit__',
                          value=self._getDescription())
    page.renderer(submitbutton)


    def render_submitbutton(self, ctx, data):
        return self.submitbutton(inevow.IRequest(ctx), ctx.tag)


    def render_liveFragment(self, ctx, data):
        return self.liveElement(inevow.IRequest(ctx), ctx.tag)


    def form(self, request, tag):
        """
        Render the inputs for a form.

        @param tag: A tag with:
            - I{form} and I{description} slots
            - I{liveform} and I{subform} patterns, to fill the I{form} slot
                - An I{inputs} slot, to fill with parameter views
            - L{IParameterView.patternName}I{-input-container} patterns for
              each parameter type in C{self.parameters}
        """
        patterns = PatternDictionary(self.docFactory)
        inputs = []

        for parameter in self.parameters:
            view = parameter.viewFactory(parameter, None)
            if view is not None:
                view.setDefaultTemplate(
                    tag.onePattern(view.patternName + '-input-container'))
                setFragmentParent = getattr(view, 'setFragmentParent', None)
                if setFragmentParent is not None:
                    setFragmentParent(self)
                inputs.append(view)
            else:
                inputs.append(_legacySpecialCases(self, patterns, parameter))

        if self.subFormName is None:
            pattern = tag.onePattern('liveform')
        else:
            pattern = tag.onePattern('subform')

        return dictFillSlots(
            tag,
            dict(form=pattern.fillSlots('inputs', inputs),
                 description=self._getDescription()))
    page.renderer(form)


    def render_form(self, ctx, data):
        return self.form(inevow.IRequest(ctx), ctx.tag)


    def invoke(self, formPostEmulator):
        """
        Invoke my callable with input from the browser.

        @param formPostEmulator: a dict of lists of strings in a format like a
            cgi-module form post.
        """
        result = self.fromInputs(formPostEmulator)
        result.addCallback(lambda params: self.callable(**params))
        return result
    expose(invoke)


    def __call__(self, formPostEmulator):
        """
        B{Private} helper which passes through to L{invoke} to support legacy
        code passing L{LiveForm} instances to L{Parameter} instead of
        L{FormParameter}.  Do B{not} call this.
        """
        return self.invoke(formPostEmulator)


    def fromInputs(self, received):
        """
        Convert some random strings received from a browser into structured
        data, using a list of parameters.

        @param received: a dict of lists of strings, i.e. the canonical Python
            form of web form post.

        @rtype: L{Deferred}
        @return: A Deferred which will be called back with a dict mapping
            parameter names to coerced parameter values.
        """
        results = []
        for parameter in self.parameters:
            name = parameter.name.encode('ascii')
            d = maybeDeferred(parameter.fromInputs, received)
            d.addCallback(lambda value, name=name: (name, value))
            results.append(d)
        return gatherResults(results).addCallback(dict)



class LiveFormFragment(_LiveFormMixin, athena.LiveFragment):
    """
    DEPRECATED.

    @see LiveForm
    """



class LiveForm(_LiveFormMixin, athena.LiveElement):
    """
    A live form.

    Create with a callable and a list of L{Parameter}s which describe the form
    of the arguments which the callable will expect.

    @ivar callable: a callable that you can call

    @ivar parameters: a list of L{Parameter} objects describing the arguments
        which should be passed to C{callable}.
    """



class _ParameterViewMixin:
    """
    Base class providing common functionality for different parameter views.

    @type parameter: L{Parameter}
    """
    def __init__(self, parameter):
        """
        @type tag: L{nevow.stan.Tag}
        @param tag: The document template to use to render this view.
        """
        self.parameter = parameter


    def __eq__(self, other):
        """
        Define equality such other views which are instances of the same class
        as this view and which wrap the same L{Parameter} are considered equal
        to this one.
        """
        if isinstance(other, self.__class__):
            return self.parameter is other.parameter
        return False


    def __ne__(self, other):
        """
        Define inequality as the negation of equality.
        """
        return not self.__eq__(other)


    def setDefaultTemplate(self, tag):
        """
        Use the given default template.
        """
        self.docFactory = stan(tag)


    def name(self, request, tag):
        """
        Render the name of the wrapped L{Parameter} or L{ChoiceParameter} instance.
        """
        return tag[self.parameter.name]
    renderer(name)


    def label(self, request, tag):
        """
        Render the label of the wrapped L{Parameter} or L{ChoiceParameter} instance.
        """
        if self.parameter.label:
            tag[self.parameter.label]
        return tag
    renderer(label)


    def description(self, request, tag):
        """
        Render the description of the wrapped L{Parameter} instance.
        """
        if self.parameter.description is not None:
            tag[self.parameter.description]
        return tag
    renderer(description)



class RepeatedLiveFormWrapper(athena.LiveElement):
    """
    A wrapper around a L{LiveForm} which has been repeated via
    L{ListChangeParameter.asLiveForm}.

    @ivar liveForm: The repeated liveform.
    @type liveForm: L{LiveForm}

    @ivar identifier: An integer identifying this repetition.
    @type identifier: C{int}

    @ivar removable: Whether this repetition can be unrepeated/removed.
    @type removable: C{bool}
    """
    fragmentName = 'repeated-liveform'
    jsClass = u'Mantissa.LiveForm.RepeatedLiveFormWrapper'

    def __init__(self, liveForm, identifier, removable=True):
        athena.LiveElement.__init__(self)
        self.liveForm = liveForm
        self.identifier = identifier
        self.removable = removable


    def getInitialArguments(self):
        """
        Include the name of the form we're wrapping, and our original values.
        """
        return (self.liveForm.subFormName.decode('utf-8'), self.identifier)


    def realForm(self, req, tag):
        """
        Render L{liveForm}.
        """
        self.liveForm.setFragmentParent(self)
        return self.liveForm
    page.renderer(realForm)


    def removeLink(self, req, tag):
        """
        Render C{tag} if L{removable} is C{True}, otherwise return the
        empty string.
        """
        if self.removable:
            return tag
        return ''
    page.renderer(removeLink)



class _TextLikeParameterView(_ParameterViewMixin, Element):
    """
    View definition base class for L{Parameter} instances which are simple text
    inputs.
    """
    def default(self, request, tag):
        """
        Render the initial value of the wrapped L{Parameter} instance.
        """
        if self.parameter.default is not None:
            tag[self.parameter.default]
        return tag
    renderer(default)



class TextParameterView(_TextLikeParameterView):
    """
    View definition for L{Parameter} instances with type of C{TEXT_INPUT}
    """
    implements(IParameterView)
    patternName = 'text'



class PasswordParameterView(_TextLikeParameterView):
    """
    View definition for L{Parameter} instances with type of C{PASSWORD_INPUT}
    """
    implements(IParameterView)
    patternName = 'password'



class OptionView(Element):
    """
    View definition for a single choice of a L{ChoiceParameter}.

    @type option: L{Option}
    """
    def __init__(self, index, option, tag):
        self._index = index
        self.option = option
        self.docFactory = stan(tag)


    def __eq__(self, other):
        """
        Define equality such other L{OptionView} instances which wrap the same
        L{Option} are considered equal to this one.
        """
        if isinstance(other, OptionView):
            return self.option is other.option
        return False


    def __ne__(self, other):
        """
        Define inequality as the negation of equality.
        """
        return not self.__eq__(other)


    def description(self, request, tag):
        """
        Render the description of the wrapped L{Option} instance.
        """
        return tag[self.option.description]
    renderer(description)


    def value(self, request, tag):
        """
        Render the value of the wrapped L{Option} instance.
        """
        return tag[self.option.value]
    renderer(value)


    def index(self, request, tag):
        """
        Render the index specified to C{__init__}.
        """
        return tag[self._index]
    renderer(index)


    def selected(self, request, tag):
        """
        Render a selected attribute on the given tag if the wrapped L{Option}
        instance is selected.
        """
        if self.option.selected:
            tag(selected='selected')
        return tag
    renderer(selected)



def _textParameterToView(parameter):
    """
    Return a L{TextParameterView} adapter for C{TEXT_INPUT}, C{PASSWORD_INPUT},
    and C{FORM_INPUT} L{Parameter} instances.
    """
    if parameter.type == TEXT_INPUT:
        return TextParameterView(parameter)
    if parameter.type == PASSWORD_INPUT:
        return PasswordParameterView(parameter)
    if parameter.type == FORM_INPUT:
        return FormInputParameterView(parameter)
    return None

registerAdapter(_textParameterToView, Parameter, IParameterView)


class ChoiceParameterView(_ParameterViewMixin, Element):
    """
    View definition for L{Parameter} instances with type of C{CHOICE_INPUT}.
    """
    implements(IParameterView)
    patternName = 'choice'

    def multiple(self, request, tag):
        """
        Render a I{multiple} attribute on the given tag if the wrapped
        L{ChoiceParameter} instance allows multiple selection.
        """
        if self.parameter.multiple:
            tag(multiple='multiple')
        return tag
    renderer(multiple)


    def options(self, request, tag):
        """
        Render each of the options of the wrapped L{ChoiceParameter} instance.
        """
        option = tag.patternGenerator('option')
        return tag[[
                OptionView(index, o, option())
                for (index, o)
                in enumerate(self.parameter.choices)]]
    renderer(options)

registerAdapter(ChoiceParameterView, ChoiceParameter, IParameterView)



class ListChangeParameterView(_ParameterViewMixin, athena.LiveElement):
    """
    L{IParameterView} adapter for L{ListChangeParameter}.

    @ivar parameter: the parameter being viewed.
    @type parameter: L{ListChangeParameter}
    """
    jsClass = u'Mantissa.LiveForm.RepeatableForm'
    patternName = 'repeatable-form'

    def __init__(self, parameter):
        self.parameter = parameter
        athena.LiveElement.__init__(self)


    def getInitialArguments(self):
        """
        Pass the name of our parameter to the client.
        """
        return (self.parameter.name,)


    def forms(self, req, tag):
        """
        Make and return some forms, using L{self.parameter.getInitialLiveForms}.

        @return: some subforms.
        @rtype: C{list} of L{LiveForm}
        """
        liveForms = self.parameter.getInitialLiveForms()
        for liveForm in liveForms:
            liveForm.setFragmentParent(self)
        return liveForms
    page.renderer(forms)


    def repeatForm(self):
        """
        Make and return a form, using L{self.parameter.asLiveForm}.

        @return: a subform.
        @rtype: L{LiveForm}
        """
        liveForm = self.parameter.asLiveForm()
        liveForm.setFragmentParent(self)
        return liveForm
    athena.expose(repeatForm)


    def repeater(self, req, tag):
        """
        Render some UI for repeating our form.
        """
        repeater = inevow.IQ(self.docFactory).onePattern('repeater')
        return repeater.fillSlots(
            'object-description', self.parameter.modelObjectDescription)
    page.renderer(repeater)

registerAdapter(ListChangeParameterView, ListChangeParameter, IParameterView)



class FormParameterView(_ParameterViewMixin, athena.LiveElement):
    """
    L{IParameterView} adapter for L{FormParameter}.

    @ivar parameter: the parameter being viewed.
    @type parameter: L{FormParameter}
    """
    implements(IParameterView)
    patternName = 'form'

    def __init__(self, parameter):
        _ParameterViewMixin.__init__(self, parameter)
        athena.LiveElement.__init__(self)


    def input(self, request, tag):
        """
        Add the wrapped form, as a subform, as a child of the given tag.
        """
        subform = self.parameter.form.asSubForm(self.parameter.name)
        subform.setFragmentParent(self)
        return tag[subform]
    renderer(input)

registerAdapter(FormParameterView, FormParameter, IParameterView)



class FormInputParameterView(_ParameterViewMixin, athena.LiveElement):
    """
    L{IParameterView} adapter for C{FORM_INPUT} L{Parameter}s.

    @ivar parameter: the parameter being viewed.
    @type parameter: L{Parameter}
    """
    implements(IParameterView)
    patternName = 'form'

    def __init__(self, parameter):
        _ParameterViewMixin.__init__(self, parameter)
        athena.LiveElement.__init__(self)


    def input(self, request, tag):
        """
        Add the wrapped form, as a subform, as a child of the given tag.
        """
        subform = self.parameter.coercer.asSubForm(self.parameter.name)
        subform.setFragmentParent(self)
        return tag[subform]
    renderer(input)



from axiom.test.historic.stubloader import saveStub
from xmantissa.people import PhoneNumber, Person

def createDatabase(s):
    PhoneNumber(store=s,
                number=u'555-1212',
                type=u'default',
                person=Person(store=s,
                              name=u'Bob'))

if __name__ == '__main__':
    saveStub(createDatabase, 7052)



from twisted.python.components import registerAdapter
from axiom.attributes import reference
from axiom.item import Item
from nevow.page import Element
from xmantissa.ixmantissa import INavigableElement, INavigableFragment
from xmantissa.webnav import Tab
from zope.interface import implements, Interface

class ISuspender(Interface):
    """
    Marker interface for suspended powerup facades.
    """

class SuspendedNavigableElement(Item):
    implements(INavigableElement, ISuspender)
    powerupInterfaces = (INavigableElement, ISuspender)

    originalNE = reference()

    def getTabs(self):
        origTabs = self.originalNE.getTabs()
        def proxyTabs(tabs):
            for tab in tabs:
                yield Tab(tab.name, self.storeID, tab.priority,
                          proxyTabs(tab.children),
                          authoritative=tab.authoritative,
                          linkURL=tab.linkURL)

        return proxyTabs(origTabs)

class SuspendedFragment(Element):
    """
    Temporary account-suspended fragment.
    """
    fragmentName = 'suspend'
    live = False
    implements(INavigableFragment)

    def head(self):
        pass

registerAdapter(SuspendedFragment, SuspendedNavigableElement, INavigableFragment)



def suspendJustTabProviders(installation):
    """
    Replace INavigableElements with facades that indicate their suspension.
    """
    if installation.suspended:
        raise RuntimeError("Installation already suspended")
    powerups = list(installation.allPowerups)
    for p in powerups:
        if INavigableElement.providedBy(p):
            p.store.powerDown(p, INavigableElement)
            sne = SuspendedNavigableElement(store=p.store, originalNE=p)
            p.store.powerUp(sne, INavigableElement)
            p.store.powerUp(sne, ISuspender)
    installation.suspended = True

def unsuspendTabProviders(installation):
    """
    Remove suspension facades and replace them with their originals.
    """
    if not installation.suspended:
        raise RuntimeError("Installation not suspended")
    powerups = list(installation.allPowerups)
    allSNEs = list(powerups[0].store.powerupsFor(ISuspender))
    for p in powerups:
        for sne in allSNEs:
            if sne.originalNE is p:
                p.store.powerDown(sne, INavigableElement)
                p.store.powerDown(sne, ISuspender)
                p.store.powerUp(p, INavigableElement)
                sne.deleteFromStore()
    installation.suspended = False
